"""
This module contains the :class:`~integreat_cms.cms.models.media.media_file.MediaFile` model as well as the helper functions
:func:`~integreat_cms.cms.models.media.media_file.upload_path` and :func:`~integreat_cms.cms.models.media.media_file.upload_path_thumbnail` which
are used to determine the file system path to which the files should be uploaded.
"""
import logging
from os.path import splitext
from time import strftime

from django.conf import settings
from django.db import models
from django.db.models import Q
from django.utils import timezone
from django.utils.formats import localize
from django.utils.translation import ugettext_lazy as _
from django.template.defaultfilters import filesizeformat


from ...constants import allowed_media
from ..abstract_base_model import AbstractBaseModel
from ..regions.region import Region
from .directory import Directory

logger = logging.getLogger(__name__)


def upload_path(instance, filename):
    """
    This function calculates the path for a file which is uploaded to the media library.
    It contains the region id, the current year and month as subdirectories and the filename.
    It is just the provisionally requested path for the media storage.
    If it already exists, Django will automatically append a random string to make sure the file name is unique.

    :param instance: The media library object
    :type instance: ~integreat_cms.cms.models.media.media_file.MediaFile

    :param filename: The filename of media library object
    :type filename: str

    :return: The upload path of the file
    :rtype: str
    """
    # If the instance already exists in the database, make sure the upload path doesn't change
    if instance.id:
        original_instance = MediaFile.objects.get(id=instance.id)
        if original_instance.file:
            logger.debug(
                "%r already exists in the database, keeping original upload path: %r",
                instance,
                original_instance.file.name,
            )
            return original_instance.file.name

    # If the media file is uploaded to a specific region, prepend a region id subdirectory
    subdirectory = f"regions/{instance.region.id}" if instance.region else "global"
    # Calculate the remaining upload path
    path = f"{subdirectory}/{strftime('%Y/%m')}/{filename}"
    logger.debug("Upload path for media file %r: %r", instance.file, path)
    return path


# pylint: disable=unused-argument
def upload_path_thumbnail(instance, filename):
    """
    This function derives the upload path of a thumbnail file from it's original file.
    This makes it a bit easier to determine which thumbnail belongs to which file if there are multiple files with the
    same file name (in which case Django automatically appends a random string to the original's file name).

    E.g. if the files ``A.jpg`` and ``A_thumbnail.jpg`` already exist, and ``A.jpg`` is uploaded again, the resulting
    file names might be e.g. ``A_EOHRFQ2.jpg`` and ``A_thumbnail_IWxPiOq.jpg``, while with this function, the thumbnail
    will be stored as ``A_EOHRFQ2_thumbnail.jpg``, making it easier to examine these files on the file system.

    :param instance: The media library object
    :type instance: ~integreat_cms.cms.models.media.media_file.MediaFile

    :param filename: The (unused) initial filename of thumbnail
    :type filename: str

    :return: The upload path of the thumbnail
    :rtype: str
    """
    # If the instance already exists in the database, make sure the upload path doesn't change
    if instance.id:
        original_instance = MediaFile.objects.get(id=instance.id)
        if original_instance.thumbnail:
            logger.debug(
                "%r already exists in the database, keeping original thumbnail upload path: %r",
                instance,
                original_instance.thumbnail.name,
            )
            return original_instance.thumbnail.name

    # Derive the thumbnail name from the original file name
    name, extension = splitext(instance.file.name)
    path = f"{name}_thumbnail{extension}"
    logger.debug("Upload path for thumbnail of %r: %r", instance.thumbnail, path)
    return path


class MediaFile(AbstractBaseModel):
    """
    The MediaFile model is used to store basic information about files which are uploaded to the CMS. This is only a
    virtual document and does not necessarily exist on the actual file system. Each document is tied to a region via its
    directory.
    """

    file = models.FileField(
        upload_to=upload_path,
        verbose_name=_("file"),
        max_length=512,
    )
    thumbnail = models.FileField(
        upload_to=upload_path_thumbnail,
        verbose_name=_("thumbnail file"),
        max_length=512,
    )
    file_size = models.IntegerField(verbose_name=_("file size"))
    type = models.CharField(
        choices=allowed_media.CHOICES,
        max_length=128,
        verbose_name=_("file type"),
    )
    name = models.CharField(max_length=512, verbose_name=_("name"))
    parent_directory = models.ForeignKey(
        Directory,
        related_name="files",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        verbose_name=_("parent directory"),
    )
    region = models.ForeignKey(
        Region,
        related_name="files",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        verbose_name=_("region"),
    )
    alt_text = models.CharField(
        max_length=512, blank=True, verbose_name=_("description")
    )
    uploaded_date = models.DateTimeField(
        auto_now_add=True,
        verbose_name=_("uploaded date"),
        help_text=_("The date and time when the media file was uploaded"),
    )
    last_modified = models.DateTimeField(
        verbose_name=_("last modified"),
        help_text=_("The date and time when the physical media file was last modified"),
    )

    @property
    def url(self):
        """
        Returns the path of the physical file

        :return: The path of the file
        :rtype: str
        """
        return settings.BASE_URL + self.file.url if self.file else None

    @property
    def thumbnail_url(self):
        """
        Return the path of the image that should be used as the thumbnail.

        :return: The path of the file. If the file is an image and no thumbnail could be generated the file itself will be returned.
        :rtype: str
        """
        if not self.thumbnail:
            if self.type.startswith("image"):
                #: Returns the path to the file itself
                return self.url
            return None
        return (
            f"{settings.BASE_URL}{self.thumbnail.url}?{self.last_modified.timestamp()}"
        )

    def serialize(self):
        """
        This methods creates a serialized string of that document. This can later be used in the AJAX calls.

        :return: A serialized string representation of the document for JSON concatenation
        :rtype: str
        """

        return {
            "id": self.id,
            "name": self.name,
            "altText": self.alt_text,
            "type": self.type,
            "typeDisplay": self.get_type_display(),
            "thumbnailUrl": self.thumbnail_url,
            "url": self.url,
            "fileSize": filesizeformat(self.file_size),
            "uploadedDate": localize(timezone.localtime(self.uploaded_date)),
            "isGlobal": not self.region,
        }

    @classmethod
    def search(cls, region, query):
        """
        Searches for all media files which match the given `query` in their name.

        :param region: The searched region
        :type region: ~integreat_cms.cms.models.regions.region.Region

        :param query: The query string used for filtering the media file
        :type query: str

        :return: A queryset for all matching objects
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.media.media_file.MediaFile ]
        """
        return cls.objects.filter(
            Q(region=region) | Q(region__isnull=True),
            Q(name__icontains=query)
            | Q(alt_text__icontains=query)
            | Q(file__icontains=query),
        )

    def __str__(self):
        """
        This overwrites the default Django :meth:`~django.db.models.Model.__str__` method which would return ``MediaFile object (id)``.
        It is used in the Django admin backend and as label for ModelChoiceFields.

        :return: A readable string representation of the document
        :rtype: str
        """
        return self.name

    def get_repr(self):
        """
        This overwrites the default Django ``__repr__()`` method which would return ``<MediaFile: MediaFile object (id)>``.
        It is used for logging.

        :return: The canonical string representation of the document
        :rtype: str
        """
        file_path = f"path: {self.file.path}, " if self.file else ""
        return (
            f"<MediaFile (id: {self.id}, name: {self.name}, {file_path}{self.region})>"
        )

    class Meta:
        #: The verbose name of the model
        verbose_name = _("media file")
        #: The plural verbose name of the model
        verbose_name_plural = _("media files")
        #: The fields which are used to sort the returned objects of a QuerySet
        ordering = ["-region", "name"]
        #: The default permissions for this model
        default_permissions = ("change", "delete", "view")
        #: The custom permissions for this model
        permissions = (
            ("upload_mediafile", "Can upload media file"),
            ("replace_mediafile", "Can replace media file"),
        )



"""
This module contains the states a translation can have
"""
from django.utils.translation import ugettext_lazy as _

#: Up to date
UP_TO_DATE = "UP_TO_DATE"
#: In translation
IN_TRANSLATION = "IN_TRANSLATION"
#: Outdated
OUTDATED = "OUTDATED"
#: Fallback
FALLBACK = "FALLBACK"
#: Missing
MISSING = "MISSING"

#: Choices to use these constants in a database field
CHOICES = (
    (UP_TO_DATE, _("Translation up-to-date")),
    (IN_TRANSLATION, _("Currently in translation")),
    (OUTDATED, _("Translation outdated")),
    # Do not show fallback translations in translation coverage
    # (FALLBACK, _("Default language is duplicated")),
    (MISSING, _("Translation missing")),
)

#: Maps from the translation state to the color used to render this state in the translation coverage view
COLORS = {
    UP_TO_DATE: "#4ade80",
    IN_TRANSLATION: "#60a5fa",
    OUTDATED: "#facc15",
    # Do not show fallback translations in translation coverage
    # FALLBACK: "#60a5fa",
    MISSING: "#f87171",
}



"""
This module contains all string representations of a month's week, used by :class:`~integreat_cms.cms.models.events.event.Event`.
"""
from django.utils.translation import ugettext_lazy as _


#: First week of the month
FIRST = 1
#: Second week of the month
SECOND = 2
#: Third week of the month
THIRD = 3
#: Fourth week of the month
FOURTH = 4
#: Last week of the month (either 4th or 5th)
LAST = 5

#: Choices to use these constants in a database field
CHOICES = (
    (FIRST, _("First week")),
    (SECOND, _("Second week")),
    (THIRD, _("Third week")),
    (FOURTH, _("Fourth week")),
    (LAST, _("Last week")),
)



"""
This module contains custom strings which are not picked up by the normal makemessages command
"""
from django.utils.translation import ugettext_lazy as _


_("Easy German")
_("German (easy)")
_("Amharic")



"""
Configuration of XLIFF app
"""
from django.apps import AppConfig


class XLIFFConfig(AppConfig):
    """
    XLIFF config inheriting the django AppConfig
    """

    name = "integreat_cms.xliff"



"""
This is a collection of tags and filters which are useful for all models.
"""
import logging

from django import template

logger = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag
def get_model_verbose_name(instance):
    """
    This tag returns the readable name of an instance's model.

    :param instance: The model object instance
    :type instance: ~django.db.models.Model

    :return: The verbose name of the model
    :rtype: str
    """
    return instance._meta.verbose_name



"""
This module contains helpers for multi-factor-authentication tasks.
"""

import random
import string


def generate_challenge(challenge_len):
    """
    This function generates a random challenge of the given length. It consists of ascii letters and digits.

    Example usage: :class:`~integreat_cms.cms.views.settings.mfa.get_mfa_challenge_view.GetMfaChallengeView`

    :param challenge_len: The desired length of the challenge
    :type challenge_len: int

    :return: A random string of the requested length
    :rtype: str
    """
    return "".join(
        [
            random.SystemRandom().choice(string.ascii_letters + string.digits)
            for i in range(challenge_len)
        ]
    )



"""
This module contains all string representations of months, used by :class:`~integreat_cms.cms.models.events.event.Event`.
"""
from django.utils.translation import ugettext_lazy as _


#: January
JANUARY = 0
#: February
FEBRUARY = 1
#: March
MARCH = 2
#: April
APRIL = 3
#: May
MAY = 4
#: June
JUNE = 5
#: July
JULY = 6
#: August
AUGUST = 7
#: September
SEPTEMBER = 8
#: October
OCTOBER = 9
#: November
NOVEMBER = 10
#: December
DECEMBER = 11

#: Choices to use these constants in a database field
CHOICES = (
    (JANUARY, _("January")),
    (FEBRUARY, _("February")),
    (MARCH, _("March")),
    (APRIL, _("April")),
    (MAY, _("May")),
    (JUNE, _("June")),
    (JULY, _("July")),
    (AUGUST, _("August")),
    (SEPTEMBER, _("September")),
    (OCTOBER, _("October")),
    (NOVEMBER, _("November")),
    (DECEMBER, _("December")),
)



"""
This modules contains the config for the view tests
"""
from django.conf import settings
from django.urls import reverse

from integreat_cms.cms.constants import status

from ...conftest import (
    ALL_ROLES,
    AUTHOR,
    EDITOR,
    HIGH_PRIV_STAFF_ROLES,
    MANAGEMENT,
    PRIV_STAFF_ROLES,
    REGION_ROLES,
    ROLES,
    ROOT,
    STAFF_ROLES,
)

#: This list contains the config for all views
#: Each element is a tuple which consists of two elements: A list of view configs and the keyword arguments that are
#: identical for all views in this list. Each view config item consists of the name of the view, the list of roles that
#: are allowed to access that view and optionally post data that is sent with the request. The post data can either be
#: a dict to send form data or a string to send JSON.
VIEWS = [
    (
        [
            ("public:login_mfa", ALL_ROLES),
            ("sitemap:index", ALL_ROLES),
            ("admin_dashboard", STAFF_ROLES),
            ("admin_feedback", STAFF_ROLES),
            ("languages", STAFF_ROLES),
            ("media_admin", STAFF_ROLES),
            ("mediacenter_directory_path", STAFF_ROLES),
            ("mediacenter_get_directory_content", STAFF_ROLES),
            ("new_language", STAFF_ROLES),
            ("new_offertemplate", STAFF_ROLES),
            ("new_region", STAFF_ROLES),
            ("new_role", [ROOT]),
            ("new_user", STAFF_ROLES),
            ("offertemplates", STAFF_ROLES),
            ("regions", STAFF_ROLES),
            ("roles", [ROOT]),
            ("user_settings", STAFF_ROLES),
            ("authenticate_modify_mfa", STAFF_ROLES),
            ("users", STAFF_ROLES),
        ],
        # The kwargs for these views
        {},
    ),
    (
        [
            ("dashboard", ROLES),
            ("languagetreenodes", STAFF_ROLES),
            ("media", ROLES),
            ("mediacenter_directory_path", ROLES),
            ("mediacenter_get_directory_content", ROLES),
            ("new_languagetreenode", STAFF_ROLES),
            (
                "new_languagetreenode",
                HIGH_PRIV_STAFF_ROLES,
                {
                    "language": 5,
                    "parent": 1,
                    "_ref_node_id": 1,
                    "_position": "first-child",
                    "visible": True,
                    "active": True,
                },
            ),
            (
                "bulk_make_languagetreenodes_visible",
                HIGH_PRIV_STAFF_ROLES,
                {"selected_ids[]": [1, 2, 3]},
            ),
            (
                "bulk_hide_languagetreenodes",
                HIGH_PRIV_STAFF_ROLES,
                {"selected_ids[]": [1, 2, 3]},
            ),
            (
                "bulk_activate_languagetreenodes",
                HIGH_PRIV_STAFF_ROLES,
                {"selected_ids[]": [1, 2, 3]},
            ),
            (
                "bulk_disable_languagetreenodes",
                HIGH_PRIV_STAFF_ROLES,
                {"selected_ids[]": [1, 2, 3]},
            ),
            ("new_region_user", STAFF_ROLES + [MANAGEMENT]),
            (
                "new_region_user",
                HIGH_PRIV_STAFF_ROLES + [MANAGEMENT],
                {"username": "new_username", "email": "new@email.address", "role": 1},
            ),
            ("region_feedback", STAFF_ROLES + [MANAGEMENT]),
            ("region_users", STAFF_ROLES + [MANAGEMENT]),
            ("translation_coverage", ROLES),
            ("organizations", STAFF_ROLES),
            ("new_organization", STAFF_ROLES),
            ("user_settings", ROLES),
            ("authenticate_modify_mfa", ROLES),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg"},
    ),
    (
        [
            ("dashboard", STAFF_ROLES),
            ("languagetreenodes", STAFF_ROLES),
            ("media", STAFF_ROLES),
            ("mediacenter_directory_path", STAFF_ROLES),
            ("mediacenter_get_directory_content", STAFF_ROLES),
            ("new_languagetreenode", STAFF_ROLES),
            (
                "new_languagetreenode",
                HIGH_PRIV_STAFF_ROLES,
                {
                    "language": 5,
                    "parent": 5,
                    "_ref_node_id": 5,
                    "_position": "first-child",
                    "visible": True,
                    "active": True,
                },
            ),
            ("new_region_user", STAFF_ROLES),
            (
                "new_region_user",
                HIGH_PRIV_STAFF_ROLES,
                {"username": "new_username", "email": "new@email.address", "role": 1},
            ),
            ("region_feedback", STAFF_ROLES),
            ("region_users", STAFF_ROLES),
            ("organizations", STAFF_ROLES),
            ("new_organization", STAFF_ROLES),
            ("translation_coverage", STAFF_ROLES),
            ("user_settings", STAFF_ROLES),
            ("authenticate_modify_mfa", STAFF_ROLES),
        ],
        # The kwargs for these views
        {"region_slug": "nurnberg"},
    ),
    (
        [
            ("sitemap:region_language", ALL_ROLES),
            ("archived_pages", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            ("archived_pois", ROLES),
            ("edit_imprint", STAFF_ROLES + [MANAGEMENT]),
            (
                "edit_imprint",
                PRIV_STAFF_ROLES + [MANAGEMENT],
                {"title": "imprint", "status": status.DRAFT},
            ),
            ("events", ROLES),
            ("events_archived", ROLES),
            ("new_event", ROLES),
            (
                "new_event",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {
                    "title": "new event",
                    "start_date": "2030-01-01",
                    "end_date": "2030-01-01",
                    "is_all_day": True,
                    "status": status.DRAFT,
                },
            ),
            ("new_page", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            (
                "new_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {
                    "title": "new page",
                    "mirrored_page_region": "",
                    "_ref_node_id": 1,
                    "_position": "first-child",
                    "status": status.DRAFT,
                },
            ),
            (
                "new_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {
                    "title": "new page",
                    "mirrored_page_region": "",
                    "_ref_node_id": 1,
                    "_position": "first-child",
                    "status": status.PUBLIC,
                },
            ),
            ("new_poi", ROLES),
            (
                "new_poi",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {
                    "title": "new poi",
                    "short_description": "short description",
                    "address": "Test-Straße 5",
                    "postcode": "54321",
                    "city": "Augsburg",
                    "country": "Deutschland",
                    "longitude": 1,
                    "latitude": 1,
                    "status": status.DRAFT,
                },
            ),
            ("pages", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            ("pois", ROLES),
            (
                "bulk_archive_pages",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"selected_ids[]": [1, 2, 3]},
            ),
            (
                "bulk_restore_pages",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"selected_ids[]": [1, 2, 3]},
            ),
            (
                "bulk_archive_events",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {"selected_ids[]": [1]},
            ),
            (
                "bulk_restore_events",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {"selected_ids[]": [1]},
            ),
            (
                "bulk_archive_pois",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {"selected_ids[]": [4]},
            ),
            (
                "bulk_restore_pois",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {"selected_ids[]": [4]},
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de"},
    ),
    (
        [
            ("archived_pages", STAFF_ROLES),
            ("archived_pois", STAFF_ROLES),
            ("edit_imprint", STAFF_ROLES),
            (
                "edit_imprint",
                PRIV_STAFF_ROLES,
                {"title": "imprint", "status": status.DRAFT},
            ),
            ("events", STAFF_ROLES),
            ("events_archived", STAFF_ROLES),
            ("new_event", STAFF_ROLES),
            (
                "new_event",
                PRIV_STAFF_ROLES,
                {
                    "title": "new event",
                    "start_date": "2030-01-01",
                    "end_date": "2030-01-01",
                    "is_all_day": True,
                    "status": status.DRAFT,
                },
            ),
            ("new_page", STAFF_ROLES),
            (
                "new_page",
                PRIV_STAFF_ROLES,
                {
                    "title": "new page",
                    "mirrored_page_region": "",
                    "_ref_node_id": 7,
                    "_position": "first-child",
                    "status": status.DRAFT,
                },
            ),
            ("new_poi", STAFF_ROLES),
            (
                "new_poi",
                PRIV_STAFF_ROLES,
                {
                    "title": "new poi",
                    "short_description": "short description",
                    "address": "Test-Straße 5",
                    "postcode": "54321",
                    "city": "Augsburg",
                    "country": "Deutschland",
                    "longitude": 1,
                    "latitude": 1,
                    "status": status.DRAFT,
                },
            ),
            ("pages", STAFF_ROLES),
            ("pois", STAFF_ROLES),
        ],
        # The kwargs for these views
        {"region_slug": "nurnberg", "language_slug": "de"},
    ),
    (
        [
            ("edit_region", STAFF_ROLES),
            (
                "edit_region",
                PRIV_STAFF_ROLES,
                {
                    "administrative_division": "CITY",
                    "name": "Augsburg",
                    "admin_mail": "augsburg@example.com",
                    "postal_code": "86150",
                    "status": "ACTIVE",
                    "longitude": 1,
                    "latitude": 1,
                    "timezone": "Europe/Berlin",
                },
            ),
        ],
        # The kwargs for these views
        {"slug": "augsburg"},
    ),
    (
        [
            ("edit_language", STAFF_ROLES),
            (
                "edit_language",
                HIGH_PRIV_STAFF_ROLES,
                {
                    "native_name": "New name",
                    "english_name": "German",
                    "slug": "de",
                    "bcp47_tag": "de-de",
                    "text_direction": "LEFT_TO_RIGHT",
                    "table_of_contents": "Inhaltsverzeichnis",
                    "primary_country_code": "de",
                    "message_content_not_available": "Foo1",
                    "message_partial_live_content_not_available": "Foo2",
                },
            ),
        ],
        # The kwargs for these views
        {"slug": "de"},
    ),
    (
        [("edit_user", STAFF_ROLES)],
        # The kwargs for these views
        {"user_id": 1},
    ),
    (
        [("edit_role", [ROOT])],
        # The kwargs for these views
        {"role_id": 1},
    ),
    (
        [("edit_offertemplate", STAFF_ROLES)],
        # The kwargs for these views
        {"slug": "ihk-lehrstellenboerse"},
    ),
    (
        [("linkcheck", ROLES)],
        # The kwargs for these views
        {"region_slug": "augsburg", "url_filter": "valid"},
    ),
    (
        [("linkcheck", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "url_filter": "valid"},
    ),
    (
        [("linkcheck", ROLES)],
        # The kwargs for these views
        {"region_slug": "augsburg", "url_filter": "unchecked"},
    ),
    (
        [("linkcheck", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "url_filter": "unchecked"},
    ),
    (
        [("linkcheck", ROLES)],
        # The kwargs for these views
        {"region_slug": "augsburg", "url_filter": "ignored"},
    ),
    (
        [("linkcheck", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "url_filter": "ignored"},
    ),
    (
        [("linkcheck", ROLES)],
        # The kwargs for these views
        {"region_slug": "augsburg", "url_filter": "invalid"},
    ),
    (
        [("linkcheck", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "url_filter": "invalid"},
    ),
    (
        [("get_page_order_table_ajax", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR])],
        # The kwargs for these views
        {"region_slug": "augsburg", "parent_id": 1},
    ),
    (
        [("get_page_order_table_ajax", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "parent_id": 7},
    ),
    (
        [
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"revision": 1, "status": status.REVIEW},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.PUBLIC},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.DRAFT},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 2, "status": status.PUBLIC},
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "page_id": 1},
    ),
    (
        [
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"revision": 1, "status": status.REVIEW},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.PUBLIC},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.DRAFT},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 2, "status": status.PUBLIC},
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "page_id": 3},
    ),
    (
        [
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"revision": 1, "status": status.REVIEW},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.PUBLIC},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.DRAFT},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 2, "status": status.PUBLIC},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 2},
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "page_id": 14},
    ),
    (
        [
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"revision": 1, "status": status.REVIEW},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.PUBLIC},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 1, "status": status.DRAFT},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 2, "status": status.PUBLIC},
            ),
            (
                "page_revisions",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {"revision": 2},
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "page_id": 15},
    ),
    (
        [("get_page_order_table_ajax", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR])],
        # The kwargs for these views
        {"region_slug": "augsburg", "parent_id": 1, "page_id": 2},
    ),
    (
        [("get_page_order_table_ajax", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "parent_id": 7, "page_id": 8},
    ),
    (
        [("get_page_order_table_ajax", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR])],
        # The kwargs for these views
        {"region_slug": "augsburg", "page_id": 2},
    ),
    (
        [("get_page_order_table_ajax", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "page_id": 8},
    ),
    (
        [("get_page_tree_ajax", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR])],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "tree_id": 2},
    ),
    (
        [("get_page_tree_ajax", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "language_slug": "de", "tree_id": 1},
    ),
    (
        [
            ("edit_page", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            ("edit_page", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            (
                "edit_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {
                    "title": "new title",
                    "mirrored_page_region": "",
                    "_ref_node_id": 21,
                    "_position": "first-child",
                    "status": status.REVIEW,
                },
            ),
            (
                "edit_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {
                    "title": "new title",
                    "mirrored_page_region": "",
                    "_ref_node_id": 21,
                    "_position": "first-child",
                    "status": status.DRAFT,
                },
            ),
            (
                "edit_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {
                    "title": "new title",
                    "mirrored_page_region": "",
                    "_ref_node_id": 24,  # Archived ref node
                    "_position": "right",
                    "status": status.DRAFT,
                },
            ),
            (
                "edit_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {
                    "title": "new title",
                    "mirrored_page_region": "",
                    "_ref_node_id": 21,
                    "_position": "first-child",
                    "status": status.PUBLIC,
                },
            ),
            ("sbs_edit_page", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            ("page_revisions", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            (
                "archive_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"post_data": True},
            ),
            (
                "restore_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"post_data": True},
            ),
            ("delete_page", HIGH_PRIV_STAFF_ROLES, {"post_data": True}),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "en", "page_id": 1},
    ),
    (
        [
            ("edit_page", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            ("edit_page", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
            (
                "edit_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {
                    "title": "new title",
                    "mirrored_page_region": "",
                    "_ref_node_id": 3,
                    "_position": "first-child",
                    "status": status.REVIEW,
                },
            ),
            (
                "edit_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {
                    "title": "new title",
                    "mirrored_page_region": "",
                    "_ref_node_id": 3,
                    "_position": "first-child",
                    "status": status.DRAFT,
                },
            ),
            (
                "edit_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR],
                {
                    "title": "new title",
                    "mirrored_page_region": "",
                    "_ref_node_id": 3,
                    "_position": "first-child",
                    "status": status.PUBLIC,
                },
            ),
            ("sbs_edit_page", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR]),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "en", "page_id": 2},
    ),
    (
        [
            ("edit_page", STAFF_ROLES),
            ("sbs_edit_page", STAFF_ROLES),
            ("page_revisions", STAFF_ROLES),
            ("archive_page", PRIV_STAFF_ROLES, {"post_data": True}),
            ("restore_page", PRIV_STAFF_ROLES, {"post_data": True}),
            ("delete_page", HIGH_PRIV_STAFF_ROLES, {"post_data": True}),
        ],
        # The kwargs for these views
        {"region_slug": "nurnberg", "language_slug": "en", "page_id": 7},
    ),
    (
        [
            (
                "move_page",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"post_data": True},
            )
        ],
        # The kwargs for these views
        {
            "region_slug": "augsburg",
            "language_slug": "en",
            "page_id": 1,
            "target_id": 21,
            "position": "first-child",
        },
    ),
    (
        [("page_revisions", STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR])],
        # The kwargs for these views
        {
            "region_slug": "augsburg",
            "language_slug": "de",
            "page_id": 1,
            "selected_revision": 1,
        },
    ),
    (
        [
            ("edit_event", ROLES),
            (
                "edit_event",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {
                    "title": "new title",
                    "start_date": "2030-01-01",
                    "end_date": "2030-01-01",
                    "is_all_day": True,
                    "status": status.DRAFT,
                },
            ),
            ("archive_event", PRIV_STAFF_ROLES + REGION_ROLES, {"post_data": True}),
            ("restore_event", PRIV_STAFF_ROLES + REGION_ROLES, {"post_data": True}),
            ("delete_event", HIGH_PRIV_STAFF_ROLES, {"post_data": True}),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "event_id": 1},
    ),
    (
        [
            ("edit_poi", ROLES),
            (
                "edit_poi",
                PRIV_STAFF_ROLES + REGION_ROLES,
                {
                    "title": "new title",
                    "short_description": "short description",
                    "address": "Test-Straße 5",
                    "postcode": "54321",
                    "city": "Augsburg",
                    "country": "Deutschland",
                    "longitude": 1,
                    "latitude": 1,
                    "status": status.DRAFT,
                },
            ),
            ("archive_poi", PRIV_STAFF_ROLES + REGION_ROLES, {"post_data": True}),
            ("restore_poi", PRIV_STAFF_ROLES + REGION_ROLES, {"post_data": True}),
            ("delete_poi", HIGH_PRIV_STAFF_ROLES, {"post_data": True}),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "poi_id": 4},
    ),
    (
        [
            ("edit_languagetreenode", STAFF_ROLES),
            (
                "edit_languagetreenode",
                HIGH_PRIV_STAFF_ROLES,
                {
                    "language": 3,
                    "parent": 2,
                    "_ref_node_id": 2,
                    "_position": "first-child",
                },
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "pk": 3},
    ),
    (
        [
            ("edit_region_user", STAFF_ROLES + [MANAGEMENT]),
            (
                "edit_region_user",
                HIGH_PRIV_STAFF_ROLES + [MANAGEMENT],
                {"username": "new_username", "email": "new@email.address", "role": 1},
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "user_id": 2},
    ),
    (
        [("edit_region_user", STAFF_ROLES)],
        # The kwargs for these views
        {"region_slug": "nurnberg", "user_id": 2},
    ),
    (
        [("dismiss_tutorial", ROLES, {"post_data": True})],
        # The kwargs for these views
        {"region_slug": "augsburg", "slug": "page-tree"},
    ),
    (
        [
            (
                "cancel_translation_process_ajax",
                PRIV_STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                {"post_data": True},
            )
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "en", "page_id": 1},
    ),
]

if settings.FCM_ENABLED:
    VIEWS += [
        (
            [
                ("new_push_notification", STAFF_ROLES + [MANAGEMENT]),
                (
                    "new_push_notification",
                    PRIV_STAFF_ROLES + [MANAGEMENT],
                    {
                        "form-TOTAL_FORMS": 2,
                        "form-INITIAL_FORMS": 0,
                        "form-MIN_NUM_FORMS": 1,
                        "form-MAX_NUM_FORMS": 2,
                        "form-0-title": "Test title",
                        "form-0-text": "Test content",
                        "form-0-language": 1,
                        "form-1-title": "",
                        "form-1-text": "",
                        "form-1-language": 6,
                        "channel": "news",
                        "mode": "ONLY_AVAILABLE",
                    },
                ),
                ("push_notifications", STAFF_ROLES + [MANAGEMENT]),
            ],
            # The kwargs for these views
            {"region_slug": "augsburg", "language_slug": "de"},
        ),
        (
            [
                ("new_push_notification", STAFF_ROLES),
                ("push_notifications", STAFF_ROLES),
            ],
            # The kwargs for these views
            {"region_slug": "nurnberg", "language_slug": "de"},
        ),
    ]

#: In order for these views to be used as parameters, we have to flatten the nested structure
PARAMETRIZED_VIEWS = [
    (view_name, kwargs, post_data[0] if post_data else {}, roles)
    for view_conf, kwargs in VIEWS
    for view_name, roles, *post_data in view_conf
]

#: This list contains the config for all views which should check whether they correctly redirect to another url
REDIRECT_VIEWS = [
    (
        [
            ("public:login", ROLES, settings.LOGIN_REDIRECT_URL),
            ("public:password_reset", ROLES, settings.LOGIN_REDIRECT_URL),
            ("public:wiki_redirect", ALL_ROLES, settings.WIKI_URL),
            ("get_mfa_challenge", STAFF_ROLES, reverse("authenticate_modify_mfa")),
            ("register_new_mfa_key", STAFF_ROLES, reverse("authenticate_modify_mfa")),
        ],
        # The kwargs for these views
        {},
    ),
    (
        [
            (
                "statistics",
                ROLES,
                reverse("dashboard", kwargs={"region_slug": "augsburg"}),
            ),
            (
                "get_mfa_challenge",
                ROLES,
                reverse("authenticate_modify_mfa", kwargs={"region_slug": "augsburg"}),
            ),
            (
                "register_new_mfa_key",
                ROLES,
                reverse("authenticate_modify_mfa", kwargs={"region_slug": "augsburg"}),
            ),
            (
                "linkcheck_landing",
                STAFF_ROLES,
                reverse(
                    "linkcheck",
                    kwargs={"region_slug": "augsburg", "url_filter": "invalid"},
                ),
            ),
        ],
        # The kwargs for these views
        {"region_slug": "augsburg"},
    ),
    (
        [
            (
                "sbs_edit_page",
                STAFF_ROLES + [MANAGEMENT, EDITOR, AUTHOR],
                reverse(
                    "edit_page",
                    kwargs={
                        "region_slug": "augsburg",
                        "language_slug": "de",
                        "page_id": 1,
                    },
                ),
            )
        ],
        # The kwargs for these views
        {"region_slug": "augsburg", "language_slug": "de", "page_id": 1},
    ),
    (
        [
            (
                "public:expand_page_translation_id",
                ALL_ROLES,
                "https://integreat.app/augsburg/de/willkommen/",
            )
        ],
        # The kwargs for these views
        {"short_url_id": 1},
    ),
]

#: In order for these views to be used as parameters, we have to flatten the nested structure
PARAMETRIZED_REDIRECT_VIEWS = [
    (view_name, kwargs, roles, target)
    for view_conf, kwargs in REDIRECT_VIEWS
    for view_name, roles, target in view_conf
]

#: Public views that only work for anonymous users
PARAMETRIZED_PUBLIC_VIEWS = [
    ("public:login", {}),
    ("public:login_mfa", {}),
    ("public:password_reset", {}),
    ("public:password_reset", {"email": "root@root.root"}),
]



import pytest

from django.forms.models import model_to_dict
from django.urls import reverse

from integreat_cms.cms.constants import status
from integreat_cms.cms.models import Region, LanguageTreeNode, Page

# pylint: disable=unused-argument,too-many-locals
@pytest.mark.django_db
def test_duplicate_regions(load_test_data, admin_client):
    """
    Test whether duplicating regions works as expected

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: tuple

    :param admin_client: The fixture providing the logged in admin
    :type admin_client: :fixture:`admin_client`
    """
    url = reverse("new_region")
    response = admin_client.post(
        url,
        data={
            "administrative_division": "CITY",
            "name": "cloned",
            "admin_mail": "cloned@example.com",
            "postal_code": "11111",
            "status": "ACTIVE",
            "longitude": 1,
            "latitude": 1,
            "duplicated_region": 1,
            "timezone": "Europe/Berlin",
        },
    )
    print(response.headers)
    assert response.status_code == 302

    source_region = Region.objects.get(id=1)
    target_region = Region.objects.get(slug="cloned")

    # Check if all cloned pages exist and are identical
    source_pages = source_region.pages.all()
    target_pages = target_region.pages.all()
    assert len(source_pages) == len(target_pages)
    for source_page, target_page in zip(source_pages, target_pages):
        source_page_dict = model_to_dict(
            source_page, exclude=["id", "tree_id", "region", "parent", "api_token"]
        )
        target_page_dict = model_to_dict(
            target_page, exclude=["id", "tree_id", "region", "parent", "api_token"]
        )
        assert source_page_dict == target_page_dict

        # Check if all cloned page translations exist and are identical
        source_page_translations = source_page.translations.all()
        target_pages_translations = target_page.translations.all()
        assert len(source_page_translations) == len(target_pages_translations)
        for source_page_translation, target_page_translation in zip(
            source_page_translations, target_pages_translations
        ):
            source_page_translation_dict = model_to_dict(
                source_page_translation, exclude=["id", "page", "status"]
            )
            target_page_translation_dict = model_to_dict(
                target_page_translation, exclude=["id", "page", "status"]
            )
            assert source_page_translation_dict == target_page_translation_dict
            assert target_page_translation.status == status.DRAFT

    # Check if all cloned language tree nodes exist and are identical
    source_language_tree = source_region.language_tree_nodes.all()
    target_language_tree = target_region.language_tree_nodes.all()
    assert len(source_language_tree) == len(target_language_tree)
    for source_node, target_node in zip(source_language_tree, target_language_tree):
        source_node_dict = model_to_dict(
            source_node, exclude=["id", "tree_id", "region", "parent"]
        )
        target_node_dict = model_to_dict(
            target_node, exclude=["id", "tree_id", "region", "parent"]
        )
        assert source_node_dict == target_node_dict

    # Check for duplicated tree ids
    for model in [LanguageTreeNode, Page]:
        # Make sure each combination of tree_id and lft resp. rgt only exists once
        for tree_field in ["lft", "rgt"]:
            combinations = model.objects.values_list("tree_id", tree_field)
            # Make sure there are no duplicates in this list
            assert len(combinations) == len(set(combinations))
        # Make sure each tree_id only exists in one region
        combinations = model.objects.values_list("tree_id", "region")
        tree_ids = [tree_id for tree_id, region in set(combinations)]
        assert len(tree_ids) == len(set(tree_ids))



"""
Configuration of SUMM.AI API app
"""
import logging
import os
import sys

from django.conf import settings
from django.apps import AppConfig

logger = logging.getLogger(__name__)


class SummAiApiConfig(AppConfig):
    """
    SUMM.AI API config inheriting the django AppConfig
    """

    name = "integreat_cms.summ_ai_api"

    def ready(self):
        """
        Inform about the SUMM.AI configuration
        """
        # Only check if running a server
        if "runserver" in sys.argv or "APACHE_PID_FILE" in os.environ:
            if not settings.SUMM_AI_ENABLED:
                logger.info("SUMM.AI API is disabled")
            elif settings.SUMM_AI_TEST_MODE:
                logger.info(
                    "SUMM.AI API is enabled, but in test mode. No credits get charged, but only a dummy text is returned."
                )
            elif settings.DEBUG:
                logger.info(
                    "SUMM.AI API is enabled, but in debug mode. Text is really translated and credits get charged, but user is 'testumgebung'"
                )
            else:
                logger.info("SUMM.AI API is enabled in production mode.")



"""
This module includes functions related to the languages API endpoint.
"""
from django.http import JsonResponse

from ..decorators import json_response


@json_response
# pylint: disable=unused-argument
def languages(request, region_slug):
    """
    Function to add all languages related to a region to a JSON.

    :param request: Django request
    :type request: ~django.http.HttpRequest
    :param region_slug: slug of a region
    :type region_slug: str

    :return: JSON object according to APIv3 languages endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    region = request.region

    result = [
        {
            "id": language.id,
            "code": language.slug,
            "bcp47_tag": language.bcp47_tag,
            "native_name": language.native_name,
            "dir": language.text_direction,
        }
        for language in region.visible_languages
    ]
    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays



"""
This module contains the base view for bulk actions
"""
import logging

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.http import Http404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import RedirectView
from django.views.generic.list import MultipleObjectMixin

from cacheops import invalidate_model
from ...deepl_api.utils import DeepLApi
from ...summ_ai_api.summ_ai_api_client import SummAiApiClient

logger = logging.getLogger(__name__)


class BulkActionView(PermissionRequiredMixin, MultipleObjectMixin, RedirectView):
    """
    View for executing a bulk action and redirect to a given location
    """

    #: The list of HTTP method names that this view will accept.
    #: The bulk action form uses only POST as submission method.
    http_method_names = ["post"]
    #: Whether the view requires change permissions
    require_change_permission = True
    #: Whether the translation objects should be prefetched
    prefetch_translations = False
    #: Whether the public translation objects should be prefetched
    prefetch_public_translations = False

    def get_permission_required(self):
        """
        Override this method to override the permission_required attribute.

        :return: The permissions that are required for views inheriting from this Mixin
        :rtype: ~collections.abc.Iterable
        """
        # If the bulk action performs changes to the database, require the change permission
        if self.require_change_permission:
            return (f"cms.change_{self.model._meta.model_name}",)
        # If the bulk action is just a read-view (e.g. export), require the view permission
        return (f"cms.view_{self.model._meta.model_name}",)

    def get_redirect_url(self, *args, **kwargs):
        r"""
        Retrieve url for redirection

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: url to redirect to
        :rtype: str
        """
        redirect_kwargs = {
            "region_slug": self.request.region.slug,
        }
        # If this bulk action is bound to a language url paramter, also pass this to the redirect url
        if "language_slug" in kwargs:
            redirect_kwargs["language_slug"] = kwargs["language_slug"]
        return reverse(f"{self.model._meta.model_name}s", kwargs=redirect_kwargs)

    def get_queryset(self):
        """
        Get the queryset of selected items for this bulk action

        :raises ~django.http.Http404: HTTP status 404 if no objects with the given ids exist

        :return: The QuerySet of the filtered links
        :rtype: ~django.db.models.query.QuerySet
        """
        # This workaround is necessary to enable the async tests for the SUMM.AI client
        logger.debug("request body: %s", self.request.body)
        queryset = (
            super()
            .get_queryset()
            .filter(
                region=self.request.region,
                id__in=self.request.POST.getlist("selected_ids[]"),
            )
        )
        if not queryset:
            raise Http404(f"No {self.model._meta.object_name} matches the given query.")
        if self.prefetch_translations:
            queryset = queryset.prefetch_translations()
        if self.prefetch_public_translations:
            queryset = queryset.prefetch_public_translations()
        return queryset


class BulkAutoTranslateView(BulkActionView):
    """
    Bulk action for automatically translating multiple objects
    """

    #: Whether the public translation objects should be prefetched
    prefetch_translations = True

    #: the form of this bulk action
    form = None

    def post(self, request, *args, **kwargs):
        r"""
        Translate multiple objects automatically

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The redirect
        :rtype: ~django.http.HttpResponseRedirect
        """
        if not settings.DEEPL_ENABLED:
            messages.error(request, _("Automatic translations are disabled"))
            return super().post(request, *args, **kwargs)
        # Collect the corresponding objects
        logger.debug("Automatic translation for: %r", self.get_queryset())
        deepl = DeepLApi()
        if deepl.check_availability(request, kwargs.get("language_slug")):
            deepl.deepl_translation(
                request, self.get_queryset(), kwargs.get("language_slug"), self.form
            )
        else:
            messages.warning(
                request,
                _(
                    "This language is not supported by DeepL. Please try another language"
                ),
            )

        # Let the base view handle the redirect
        return super().post(request, *args, **kwargs)


class BulkActionEasyGermanView(BulkActionView):
    """
    Bulk action for translating multiple objects to Easy German
    """

    #: the form of this bulk action
    form = None

    def post(self, request, *args, **kwargs):
        r"""
        Translate multiple objects automatically to Easy German

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The redirect
        :rtype: ~django.http.HttpResponseRedirect
        """
        if not settings.SUMM_AI_ENABLED or not request.region.summ_ai_enabled:
            if not settings.SUMM_AI_ENABLED:
                logger.warning("SUMM.AI globally disabled")
            if not request.region.summ_ai_enabled:
                logger.warning("SUMM.AI disabled in %r", request.region)
            messages.error(request, _("Translations to Easy German are disabled"))
            return super().post(request, *args, **kwargs)
        # Collect the corresponding objects
        logger.info(
            "%r started SUMM.AI translation for: %r", request.user, self.get_queryset()
        )
        api_client = SummAiApiClient(request, self.form)
        api_client.translate_queryset(self.get_queryset())
        # Let the base view handle the redirect
        return super().post(request, *args, **kwargs)


class BulkUpdateBooleanFieldView(BulkActionView):
    """
    Bulk action for toggling boolean fields of multiple objects at once
    """

    #: The value of the field (defaults to ``True``)
    value = True

    @property
    def field_name(self):
        """
        Called when the bulk action is performed and the ``field_name`` attribute was not overwritten

        :raises NotImplementedError: If the ``field_name`` attribute is not implemented in the subclass
        """
        raise NotImplementedError(
            "Subclasses of BulkUpdateBooleanFieldView must provide a 'field_name' attribute"
        )

    @property
    def action(self):
        """
        Called when the bulk action is performed and the ``action`` attribute was not overwritten

        :raises NotImplementedError: If the ``action`` attribute is not implemented in the subclass
        """
        raise NotImplementedError(
            "Subclasses of BulkUpdateBooleanFieldView must provide an 'action' attribute"
        )

    def post(self, request, *args, **kwargs):
        r"""
        Update the fields of the selected objects and redirect

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The redirect
        :rtype: ~django.http.HttpResponseRedirect
        """

        # Archive objects
        self.get_queryset().update(**{self.field_name: self.value})
        # Invalidate cache
        invalidate_model(self.model)
        logger.debug(
            "Updated %s=%s for %r by %r",
            self.field_name,
            self.value,
            self.get_queryset(),
            request.user,
        )
        messages.success(
            request,
            _("The selected {} were successfully {}").format(
                self.model._meta.verbose_name_plural, self.action
            ),
        )
        # Let the base view handle the redirect
        return super().post(request, *args, **kwargs)


# pylint: disable=too-many-ancestors
class BulkArchiveView(BulkUpdateBooleanFieldView):
    """
    Bulk action for restoring multiple objects at once
    """

    #: The name of the archived-field
    field_name = "archived"

    #: The name of the action
    action = _("archived")


# pylint: disable=too-many-ancestors
class BulkRestoreView(BulkArchiveView):
    """
    Bulk action for restoring multiple objects at once
    """

    #: The value of the field
    value = False

    #: The name of the action
    action = _("restored")



from ..custom_model_form import CustomModelForm
from ...models import OfferTemplate
from ...utils.slug_utils import generate_unique_slug_helper


class OfferTemplateForm(CustomModelForm):
    """
    Form for creating and modifying offer template objects
    """

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = OfferTemplate
        #: The fields of the model which should be handled by this form
        fields = ["name", "slug", "thumbnail", "url", "post_data", "use_postal_code"]

    def __init__(self, *args, **kwargs):
        r"""
        Initialize offer template form

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict
        """
        super().__init__(*args, **kwargs)

        self.fields["slug"].required = False

    def clean_slug(self):
        """
        Validate the slug field (see :ref:`overriding-modelform-clean-method`)

        :return: A unique slug based on the input value
        :rtype: str
        """
        return generate_unique_slug_helper(self, "offer-template")

    def clean_post_data(self):
        """
        Validate the post data field (see :ref:`overriding-modelform-clean-method`)

        :return: The valid post data
        :rtype: str
        """
        cleaned_post_data = self.cleaned_data["post_data"]
        if not cleaned_post_data:
            cleaned_post_data = {}
        return cleaned_post_data



"""
This is a collection of tags and filters which are useful for all content types (:class:`~integreat_cms.cms.models.pages.page.Page`,
:class:`~integreat_cms.cms.models.events.event.Event` and :class:`~integreat_cms.cms.models.pois.poi.POI`).
"""
import logging

from django.urls import reverse

from django import template
from django.utils.translation import gettext as _

from ..constants import translation_status
from ..models import Language, PageTranslation, EventTranslation, POITranslation

logger = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag
def get_translation(instance, language_slug):
    """
    This tag returns the most recent translation of the requested content object in the requested language.

    :param instance: The content object instance
    :type instance: ~integreat_cms.cms.models.pages.page.Page, ~integreat_cms.cms.models.events.event.Event or ~integreat_cms.cms.models.pois.poi.POI

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :return: The translation object of the requested instance
    :rtype: ~integreat_cms.cms.models.pages.page_translation.PageTranslation, ~integreat_cms.cms.models.events.event_translation.EventTranslation,
            or ~integreat_cms.cms.models.pois.poi_translation.POITranslation
    """
    return instance.get_translation(language_slug)


@register.simple_tag
def get_public_translation(instance, language_slug):
    """
    This tag returns the most recent public translation of the requested content object in the requested language.

    :param instance: The content object instance
    :type instance: ~integreat_cms.cms.models.pages.page.Page, ~integreat_cms.cms.models.events.event.Event or ~integreat_cms.cms.models.pois.poi.POI

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :return: The translation object of the requested instance
    :rtype: ~integreat_cms.cms.models.pages.page_translation.PageTranslation, ~integreat_cms.cms.models.events.event_translation.EventTranslation,
            or ~integreat_cms.cms.models.pois.poi_translation.POITranslation
    """
    return instance.get_public_translation(language_slug)


@register.simple_tag
def translated_language_name(language_slug):
    """
    This tag returns the name of the requested language in the current backend language

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :return: The translated name of the requested language
    :rtype: str
    """
    language = Language.objects.filter(slug=language_slug)
    if language.exists():
        return language.first().translated_name
    return ""


@register.simple_tag
def get_language(language_slug):
    """
    This tag returns the requested language by slug

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :return: The requested language
    :rtype: ~integreat_cms.cms.models.languages.language.Language
    """
    return Language.objects.filter(slug=language_slug).first()


@register.simple_tag
def minor_edit_label(region, language):
    """
    This tag returns the label of the minor edit field of the given form

    :param region: current region
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The current language
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :return: The minor edit label
    :rtype: str
    """
    language_node = region.language_node_by_slug[language.slug]
    if language_node.is_leaf():
        return _("Implications on the translation status")
    if language_node.is_root():
        return _("Implications on other translations")
    return _("Implications on the status of this and other translations")


@register.simple_tag
def minor_edit_help_text(region, language, translation_form):
    """
    This tag returns the help text of the minor edit field of the given form

    :param region: current region
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The current language
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param translation_form: The given model form
    :type translation_form: ~integreat_cms.cms.forms.custom_model_form.CustomModelForm

    :return: The minor edit help text
    :rtype: str
    """
    language_node = region.language_node_by_slug[language.slug]
    if language_node.is_leaf():
        return _("Tick if this edit should not change the status of this translation.")
    if language_node.is_root():
        return translation_form["minor_edit"].help_text
    return _(
        "Tick if this edit should not change the status of this translation and does not require an update of translations in other languages."
    )


@register.simple_tag
def build_url(target, region_slug, language_slug, content_field=None, content_id=None):
    """
    This tag returns the requested language by slug

    :param target: The slug of the requested target
    :type target: str

    :param region_slug: The slug of the requested region
    :type region_slug: str

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :param content_field: The name of the content field
    :type content_field: str

    :param content_id: The id of the content
    :type content_id: int

    :return: list of args
    :rtype: str
    """
    kwargs = {
        "region_slug": region_slug,
        "language_slug": language_slug,
    }
    if content_field and content_id:
        kwargs[content_field] = content_id
    return reverse(target, kwargs=kwargs)


@register.filter
def remove(elements, element):
    """
    This tag removes an element from a list.

    :param elements: The given list of elements
    :type elements: list

    :param element: The element to be removed
    :type element: object

    :return: The list without the element
    :rtype: list
    """
    try:
        # Copy input list to make sure it is not modified in place
        result = elements.copy()
        result.remove(element)
        return result
    except ValueError:
        # If the element wasn't in the list to begin with, just return the input
        return elements


@register.filter
def sort_translation_states(translation_states, second_language):
    """
    This filter sorts languages in language tabs

    :param translation_states: Other languages
    :type translation_states: dict [ tuple ( ~integreat_cms.cms.models.languages.language.Language, str ) ]

    :param second_language: The current language
    :type second_language: ~integreat_cms.cms.models.languages.language.Language

    :return: the filtered list with the current language on the second position
    :rtype: list
    """
    try:
        first_language_slug = next(iter(translation_states))
    except StopIteration:
        return [(second_language, translation_status.MISSING)]
    if first_language_slug != second_language.slug:
        second_language_state = translation_states.pop(second_language.slug)
        translation_states = list(translation_states.values())
        translation_states.insert(1, second_language_state)
    else:
        translation_states = list(translation_states.values())
    return translation_states


@register.filter
def get_int_list(data, list_name):
    """
    This filter returns the list data of a one-to-many field as ints.

    :param data: The requested form data
    :type data: ~django.http.QueryDict

    :param list_name: The name of the requested field
    :type list_name: str

    :return: The int value list
    :rtype: list [ int ]
    """
    return [int(item) for item in data.getlist(list_name)]


@register.filter
def is_empty(iterable):
    """
    This filter checks whether the given iterable is empty.

    :param iterable: The requested iterable
    :type iterable: ~collections.abc.Iterable

    :return: Whether or not the given iterable is empty
    :rtype: bool
    """
    return not bool(iterable)


@register.simple_tag
def object_translation_has_view_perm(user, obj):
    """
    This filter accepts any translation of Event, Page or Poi and returns
    whether the user has the permission to view this object

    :param user: The requested user
    :type user: ~integreat_cms.cms.models.users.user.User

    :param obj: The requested object
    :type obj: ~integreat_cms.cms.models.pages.page_translation.PageTranslation,
               ~integreat_cms.cms.models.events.event_translation.EventTranslation, or ~integreat_cms.cms.models.pois.poi_translation.POITranslation

    :raises ValueError: if the object is not a translation of Event, Page or Poi

    :return: Whether the user is allowed to view this object
    :rtype: bool
    """
    if isinstance(obj, EventTranslation):
        return user.has_perm("cms.view_event")
    if isinstance(obj, PageTranslation):
        return user.has_perm("cms.view_page")
    if isinstance(obj, POITranslation):
        return user.has_perm("cms.view_poi")
    raise ValueError(f"Invalid model: {type(obj)}")



import pytest

from django.urls import reverse

from .view_config import PARAMETRIZED_REDIRECT_VIEWS


@pytest.mark.django_db
@pytest.mark.parametrize("view_name,kwargs,roles,target", PARAMETRIZED_REDIRECT_VIEWS)
def test_view_redirect(login_role_user, view_name, kwargs, roles, target):
    """
    This test checks whether the given view redirects to the given target url

    :param login_role_user: The fixture providing the http client and the current role (see :meth:`~tests.conftest.login_role_user`)
    :type login_role_user: tuple

    :param view_name: The identifier of the view
    :type view_name: str

    :param kwargs: The keyword argument passed to the view
    :type kwargs: dict

    :param roles: The list of roles which should experience the redirect
    :type roles: list

    :param target: The expected redirection target url
    :type target: str
    """
    client, role = login_role_user
    # For redirection checks, we only test the given roles
    if role in roles:
        url = reverse(view_name, kwargs=kwargs)
        response = client.get(url)
        print(response.headers)
        assert response.status_code == 302
        assert response.headers.get("location") == target



"""
This package contains all views related to push notifications
"""
from .push_notification_list_view import PushNotificationListView
from .push_notification_form_view import PushNotificationFormView



"""
This module contains constants which represent all possible administrative divisions of a region in Germany.
"""
from django.utils.translation import ugettext_lazy as _


#: Federal state (German: "Bundesland")
FEDERAL_STATE = "FEDERAL_STATE"
#: Area state (German: "Flächenland")
AREA_STATE = "AREA_STATE"
#: Free state (German: "Freistaat")
FREE_STATE = "FREE_STATE"
#: City state (German: "Stadtstaat")
CITY_STATE = "CITY_STATE"
#: Governmental district (German: "Regierungsbezirk")
GOVERNMENTAL_DISTRICT = "GOVERNMENTAL_DISTRICT"
#: Urban district (German: "Kreisfreie Stadt")
URBAN_DISTRICT = "URBAN_DISTRICT"
#: Rural district (German: "Landkreis")
RURAL_DISTRICT = "RURAL_DISTRICT"
#: District (German: "Kreis")
DISTRICT = "DISTRICT"
#: City (German: "Stadt")
CITY = "CITY"
#: City and district (German: "Stadt und Landkreis")
CITY_AND_DISTRICT = "CITY_AND_DISTRICT"
#: Region (German: "Region")
REGION = "REGION"
#: Municipality (German: "Gemeinde")
MUNICIPALITY = "MUNICIPALITY"
#: Collective municipality (German: "Verbandsgemeinde")
COLLECTIVE_MUNICIPALITY = "COLLECTIVE_MUNICIPALITY"
#: Initial reception center (German: "Erstaufnahmeeinrichtung")
INITIAL_RECEPTION_CENTER = "INITIAL_RECEPTION_CENTER"

#: Choices to use these constants in a database field
CHOICES = (
    (CITY, _("City")),
    (DISTRICT, _("District")),
    (RURAL_DISTRICT, _("Rural district")),
    (REGION, _("Region")),
    (CITY_AND_DISTRICT, _("City and district")),
    (URBAN_DISTRICT, _("Urban district")),
    (GOVERNMENTAL_DISTRICT, _("Governmental district")),
    (CITY_STATE, _("City state")),
    (AREA_STATE, _("Area state")),
    (FREE_STATE, _("Free state")),
    (FEDERAL_STATE, _("Federal state")),
    (MUNICIPALITY, _("Municipality")),
    (COLLECTIVE_MUNICIPALITY, _("Collective municipality")),
    (INITIAL_RECEPTION_CENTER, _("Initial reception center")),
)



"""
This module contains the possible modes for push notifications.
"""
from django.utils.translation import ugettext_lazy as _


#: Send only available translations
ONLY_AVAILABLE = "ONLY_AVAILABLE"
#: Use main language if no translation is available
USE_MAIN_LANGUAGE = "USE_MAIN_LANGUAGE"

#: Choices to use these constants in a database field
PN_MODES = (
    (ONLY_AVAILABLE, _("Only send available translations")),
    (USE_MAIN_LANGUAGE, _("Use main language if no translation is available")),
)



import logging

from copy import deepcopy

from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import get_language, ugettext_lazy as _
from django.utils import timezone

from ..constants import status, translation_status
from ..utils.content_edit_lock import get_locking_user
from .regions.region import Region
from .abstract_base_model import AbstractBaseModel

logger = logging.getLogger(__name__)


class ContentQuerySet(models.QuerySet):
    """
    This queryset provides the option to prefetch translations for content objects
    """

    def prefetch_translations(self, to_attr="prefetched_translations", **filters):
        r"""
        Get the queryset including the custom attribute ``to_attr`` which contains the latest
        translations of each content object in each language, optionally filtered by the given ``status``

        :param to_attr: To which attribute the prefetched translations should be added [optional, defaults to ``prefetched_translations``]
        :type to_attr: str

        :param \**filters: Additional filters to be applied on the translations (e.g. by status)
        :type \**filters: dict

        :return: The queryset of content objects
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel ]
        """
        TranslationModel = self.model.get_translation_model()
        foreign_field = TranslationModel.foreign_field() + "_id"
        return self.prefetch_related(
            models.Prefetch(
                "translations",
                queryset=TranslationModel.objects.filter(**filters)
                .order_by(foreign_field, "language_id", "-version")
                .distinct(foreign_field, "language_id")
                .select_related("language"),
                to_attr=to_attr,
            )
        )

    def prefetch_public_translations(self):
        """
        Get the queryset including the custom attribute ``prefetched_public_translations`` which contains the latest
        public translations of each content object in each language

        :return: The queryset of content objects
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel ]
        """
        return self.prefetch_translations(
            to_attr="prefetched_public_translations", status=status.PUBLIC
        )

    def prefetch_public_or_draft_translations(self):
        """
        Get the queryset including the custom attribute ``prefetched_public_or_draft_translations`` which contains the latest
        public or draft translations of each content object in each language

        :return: The queryset of content objects
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel ]
        """
        return self.prefetch_translations(
            to_attr="prefetched_public_or_draft_translations",
            status__in=[status.DRAFT, status.PUBLIC],
        )

    def prefetch_major_public_translations(self):
        """
        Get the queryset including the custom attribute ``prefetched_major_public_translations`` which contains the
        latest major (in other words not a minor edit) public translations of each content object in each language

        :return: The queryset of content objects
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.abstract_content_model.AbstractContentModel ]
        """
        return self.prefetch_translations(
            to_attr="prefetched_major_public_translations",
            status=status.PUBLIC,
            minor_edit=False,
        )


class AbstractContentModel(AbstractBaseModel):
    """
    Abstract base class for all content models
    """

    region = models.ForeignKey(
        Region, on_delete=models.CASCADE, verbose_name=_("region")
    )
    created_date = models.DateTimeField(
        default=timezone.now, verbose_name=_("creation date")
    )

    #: Custom model manager for content objects
    objects = ContentQuerySet.as_manager()

    #: Whether translations should be returned in the default language if they do not exist
    fallback_translations_enabled = False

    @cached_property
    def languages(self):
        """
        This property returns a list of all :class:`~integreat_cms.cms.models.languages.language.Language` objects, to
        which a translation exists.

        :return: The existing languages of this content object
        :rtype: list
        """
        translations = self.prefetched_translations_by_language_slug.values()
        return [translation.language for translation in translations]

    def available_translations(self):
        """
        This method returns an iterator over all available translations, respecting the `fallback_translations_enabled` setting.

        :return: An iterator over all translations
        :rtype: Iterator[:class:`~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation`]
        """
        # Check if fallback translation should be used
        if self.fallback_translations_enabled:
            all_languages = self.region.visible_languages
        else:
            all_languages = self.public_languages

        for language in all_languages:
            public_translation = self.get_public_translation(language.slug)
            if public_translation:
                yield public_translation

    @cached_property
    def public_languages(self):
        """
        This property returns a list of all :class:`~integreat_cms.cms.models.languages.language.Language` objects, to
        which a public translation exists and which are visible in this region.

        :return: The existing languages of this content object
        :rtype: list
        """
        translations = self.prefetched_public_translations_by_language_slug.values()
        return [
            translation.language
            for translation in translations
            if translation.language in self.region.visible_languages
        ]

    def get_prefetched_translations_by_language_slug(
        self, attr="prefetched_translations", **filters
    ):
        r"""
        This method returns a mapping from language slugs to their latest translations of this object

        :param attr: Which attribute should be tried to get the prefetched translations [optional, defaults to ``"prefetched_translations"``]
        :type attr: str

        :param \**filters: Additional filters to be applied on the translations (e.g. by status)
        :type \**filters: dict

        :return: The prefetched translations by language slug
        :rtype: dict
        """
        try:
            # Try to get the prefetched translations (which are already distinct per language)
            prefetched_translations = getattr(self, attr)
        except AttributeError:
            # If the translations were not prefetched, query it from the database
            prefetched_translations = (
                self.translations.filter(**filters)
                .select_related("language")
                .order_by("language__id", "-version")
                .distinct("language__id")
                .all()
            )
        return {
            translation.language.slug: translation
            for translation in prefetched_translations
        }

    @cached_property
    def prefetched_translations_by_language_slug(self):
        """
        This method returns a mapping from language slugs to their latest translations of this object

        :return: The prefetched translations by language slug
        :rtype: dict
        """
        return self.get_prefetched_translations_by_language_slug()

    def get_translation(self, language_slug):
        """
        This function uses the reverse foreign key ``self.translations`` to get all translations of ``self``
        and filters them to the requested :class:`~integreat_cms.cms.models.languages.language.Language` slug.

        :param language_slug: The slug of the desired :class:`~integreat_cms.cms.models.languages.language.Language`
        :type language_slug: str

        :return: The translation in the requested :class:`~integreat_cms.cms.models.languages.language.Language` or
                 :obj:`None` if no translation exists
        :rtype: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
        """
        return self.prefetched_translations_by_language_slug.get(language_slug)

    @cached_property
    def prefetched_public_translations_by_language_slug(self):
        """
        This method returns a mapping from language slugs to their public translations of this object

        :return: The object translation in the requested :class:`~integreat_cms.cms.models.languages.language.Language` or
                 :obj:`None` if no translation exists
        :rtype: dict
        """
        return self.get_prefetched_translations_by_language_slug(
            attr="prefetched_public_translations", status=status.PUBLIC
        )

    def get_public_translation(self, language_slug):
        """
        This function retrieves the newest public translation of a content object.

        :param language_slug: The slug of the requested :class:`~integreat_cms.cms.models.languages.language.Language`
        :type language_slug: str

        :return: The public translation of a content object
        :rtype: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
        """
        public_translation = self.prefetched_public_translations_by_language_slug.get(
            language_slug
        )
        # Check if fallback translation should be used
        if not public_translation and self.fallback_translations_enabled:
            # Get the fallback translation
            public_translation = (
                self.prefetched_public_translations_by_language_slug.get(
                    self.region.default_language.slug
                )
            )
            if public_translation:
                # Create copy in memory to make sure original translation is not affected by changes
                public_translation = deepcopy(public_translation)
                # Reset id to make sure id does not conflict with existing translation
                public_translation.id = None
                # Fake the requested language
                public_translation.language = self.region.language_node_by_slug[
                    language_slug
                ].language
                # Reset prefetched translations
                public_translation.foreign_object.prefetched_public_translations_by_language_slug = (
                    self.prefetched_public_translations_by_language_slug
                )
                # Clear cached property in case url with different language was already calculated before
                try:
                    del public_translation.url_prefix
                except AttributeError:
                    pass
        return public_translation

    @cached_property
    def prefetched_public_or_draft_translations_by_language_slug(self):
        """
        This method returns a mapping from language slugs to their public translations of this object

        :return: The object translation in the requested :class:`~integreat_cms.cms.models.languages.language.Language` or
                 :obj:`None` if no translation exists
        :rtype: dict
        """
        return self.get_prefetched_translations_by_language_slug(
            attr="prefetched_public_or_draft_translations",
            status__in=[status.DRAFT, status.PUBLIC],
        )

    def get_public_or_draft_translation(self, language_slug):
        """
        This function retrieves the newest public or draft translation of a content object.

        :param language_slug: The slug of the requested :class:`~integreat_cms.cms.models.languages.language.Language`
        :type language_slug: str

        :return: The public translation of a content object
        :rtype: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
        """
        translation = self.prefetched_public_or_draft_translations_by_language_slug.get(
            language_slug
        )
        return translation

    @cached_property
    def prefetched_major_public_translations_by_language_slug(self):
        """
        This method returns a mapping from language slugs to their major public translations of this object

        :return: The object translation in the requested :class:`~integreat_cms.cms.models.languages.language.Language` or
                 :obj:`None` if no translation exists
        :rtype: dict
        """
        return self.get_prefetched_translations_by_language_slug(
            attr="prefetched_major_public_translations",
            status=status.PUBLIC,
            minor_edit=False,
        )

    def get_major_public_translation(self, language_slug):
        """
        This function retrieves the newest major public translation of a content object.

        :param language_slug: The slug of the requested :class:`~integreat_cms.cms.models.languages.language.Language`
        :type language_slug: str

        :return: The public translation of a content object
        :rtype: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
        """
        return self.prefetched_major_public_translations_by_language_slug.get(
            language_slug
        )

    @cached_property
    def backend_translation(self):
        """
        This function returns the translation of this content object in the current backend language.

        :return: The backend translation of a content object
        :rtype: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
        """
        return self.get_translation(get_language())

    @cached_property
    def default_translation(self):
        """
        This function returns the translation of this content object in the region's default language.
        Since a content object can only be created by creating a translation in the default language, this is guaranteed
        to return a object translation (Exception: When the default language tree node is changed to another language
        after the page has been created, the default translation might not exist).

        :return: The default translation of a content object
        :rtype: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
        """
        return self.get_translation(self.region.default_language.slug)

    @cached_property
    def best_translation(self):
        """
        This function returns the translation of this content object in the current backend language and if it doesn't
        exist, it provides a fallback to the translation in the region's default language.

        :return: The "best" translation of a content object for displaying in the backend
        :rtype: ~integreat_cms.cms.models.abstract_content_translation.AbstractContentTranslation
        """
        return (
            self.backend_translation
            or self.default_translation
            or self.translations.first()
        )

    def get_translation_state(self, language_slug):
        """
        This function returns the current state of a translation in the given language.

        :param language_slug: The slug of the desired :class:`~integreat_cms.cms.models.languages.language.Language`
        :type language_slug: str

        :return: A string describing the state of the translation, one of :data:`~integreat_cms.cms.constants.translation_status.CHOICES`
        :rtype: str
        """
        translation = self.get_major_public_translation(language_slug)
        if not translation:
            if self.fallback_translations_enabled:
                fallback_translation = self.get_public_translation(
                    self.region.default_language.slug
                )
                if fallback_translation:
                    return translation_status.FALLBACK
            return translation_status.MISSING
        return translation.translation_state

    @cached_property
    def translation_states(self):
        """
        This property calculates all translations states of the object

        :return: A dictionary containing each language as key and the given translation state as value
        :rtype: dict
        """
        return {
            node.slug: (node.language, self.get_translation_state(node.slug))
            for node in self.region.language_tree
            if node.active
        }

    @property
    def edit_lock_key(self):
        """
        This property returns the key that is used to lock this specific content object

        :return: A tuple of the id of this object and the classname
        :rtype: tuple
        """
        return (self.id, type(self).__name__)

    def get_locking_user(self):
        """
        This method returns the user that is currently locking this content object.

        :return: The user
        :rtype: ~django.contrib.auth.models.User
        """
        return get_locking_user(*self.edit_lock_key)

    def __str__(self):
        """
        This overwrites the default Django :meth:`~django.db.models.Model.__str__` method which would return ``AbstractContentModel object (id)``.
        It is used in the Django admin backend and as label for ModelChoiceFields.

        :return: A readable string representation of the content object
        :rtype: str
        """
        return self.best_translation.title

    def get_repr(self):
        """
        This overwrites the default Django ``__repr__()`` method which would return ``<AbstractContentModel: AbstractContentModel object (id)>``.
        It is used for logging.

        :return: The canonical string representation of the content object
        :rtype: str
        """
        class_name = type(self).__name__
        return f"<{class_name} (id: {self.id}, region: {self.region.slug}, slug: {self.best_translation.slug})>"

    class Meta:
        #: This model is an abstract base class
        abstract = True



import logging

from django.http import JsonResponse, Http404

from ....cms.models import POIFeedback
from ...decorators import json_response, feedback_handler

logger = logging.getLogger(__name__)


@feedback_handler
@json_response
def poi_feedback(data, region, language, comment, rating, is_technical):
    """
    Store feedback about single POI in database

    :param data: HTTP request body data
    :type data: dict

    :param region: The region of this sitemap's urls
    :type region: ~integreat_cms.cms.models.regions.region.Region

    :param language: The language of this sitemap's urls
    :type language: ~integreat_cms.cms.models.languages.language.Language

    :param comment: The comment sent as feedback
    :type comment: str

    :param rating: up or downvote, neutral
    :type rating: str

    :param is_technical: is feedback on content or on tech
    :type is_technical: bool

    :raises ~django.http.Http404: HTTP status 404 if no POI with the given slug exists.

    :return: decorated function that saves feedback in database
    :rtype: ~collections.abc.Callable
    """
    poi_translation_slug = data.get("slug")

    pois = region.pois.filter(
        translations__slug=data.get("slug"),
        translations__language=language,
    ).distinct()

    if len(pois) > 1:
        logger.error(
            "POI translation slug %r is not unique per region and language, found multiple: %r",
            poi_translation_slug,
            pois,
        )
        return JsonResponse({"error": "Internal Server Error"}, status=500)
    if len(pois) == 0:
        raise Http404("No matching location found for slug.")
    poi = pois[0]
    poi_translation = poi.get_translation(language.slug)

    POIFeedback.objects.create(
        poi_translation=poi_translation,
        region=region,
        language=language,
        rating=rating,
        comment=comment,
        is_technical=is_technical,
    )
    return JsonResponse({"success": "Feedback successfully submitted"}, status=201)



from django.contrib import messages
from django.utils.translation import gettext as _

from ..form_views import CustomUpdateView
from ...forms import RegionForm


class RegionUpdateView(CustomUpdateView):
    """
    View for updating regions
    """

    #: The form class for this update view
    form_class = RegionForm

    def get(self, request, *args, **kwargs):
        r"""
        Render region form for HTTP GET requests

        :param request: The current request
        :type request: ~django.http.HttpRequest

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """
        # Populate self.object
        response = super().get(request, *args, **kwargs)
        # Show warning when locations are enabled without bounding box
        if self.object.locations_enabled and not self.object.has_bounding_box:
            messages.warning(
                request,
                _(
                    "Locations are enabled but the bounding box coordinates are incomplete."
                ),
            )
        return response



"""
This package contains views related to login, logout and password reset functionality as well as 2FA authentication:
"""
from .account_activation_view import AccountActivationView
from .login_view import LoginView
from .logout_view import LogoutView
from .password_reset_view import PasswordResetView
from .password_reset_confirm_view import PasswordResetConfirmView

from .mfa.mfa_login_view import MfaLoginView
from .mfa.mfa_assert_view import MfaAssertView
from .mfa.mfa_verify_view import MfaVerifyView



import json
import pytest

from django.test.client import Client

from .api_config import API_ENDPOINTS


# pylint: disable=unused-argument
@pytest.mark.django_db
@pytest.mark.parametrize(
    "endpoint,wp_endpoint,expected_result,expected_code", API_ENDPOINTS
)
def test_api_result(
    load_test_data, endpoint, wp_endpoint, expected_result, expected_code
):
    """
    This test class checks all endpoints defined in :attr:`~tests.api.api_config.API_ENDPOINTS`.
    It verifies that the content delivered by the endpoint is equivalent with the data
    provided in the corresponding json file.

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: tuple

    :param endpoint: The url of the new Django pattern
    :type endpoint: str

    :param wp_endpoint: The legacy url of the wordpress endpoint pattern
    :type wp_endpoint: str

    :param expected_result: The path to the json file that contains the expected result
    :type expected_result: str

    :param expected_code: The expected HTTP status code
    :type expected_code: int
    """
    client = Client()
    response = client.get(endpoint, format="json")
    print(response.headers)
    assert response.status_code == expected_code
    response_wp = client.get(wp_endpoint, format="json")
    print(response_wp.headers)
    assert response_wp.status_code == expected_code
    with open(expected_result, encoding="utf-8") as f:
        result = json.load(f)
        assert result == response.json()
        assert result == response_wp.json()



import logging

from django.contrib import messages
from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext as _
from django.views.generic.list import MultipleObjectMixin

from ....xliff.utils import pages_to_xliff_file
from ...models import Page
from ...utils.pdf_utils import generate_pdf
from ...utils.translation_utils import ugettext_many_lazy as __
from ..bulk_action_views import BulkActionView

logger = logging.getLogger(__name__)


class PageBulkActionMixin(MultipleObjectMixin):
    """
    Mixin for page bulk actions
    """

    #: The model of this :class:`~integreat_cms.cms.views.bulk_action_views.BulkActionView`
    model = Page


# pylint: disable=too-many-ancestors
class GeneratePdfView(PageBulkActionMixin, BulkActionView):
    """
    Bulk action for generating a PDF document of the content
    """

    #: Whether the view requires change permissions
    require_change_permission = False
    #: Whether the public translation objects should be prefetched
    prefetch_public_translations = True

    def post(self, request, *args, **kwargs):
        r"""
        Apply the bulk action on every item in the queryset and redirect

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The redirect
        :rtype: ~django.http.HttpResponseRedirect
        """
        # Generate PDF document and redirect to it
        return generate_pdf(
            request.region,
            kwargs.get("language_slug"),
            self.get_queryset(),
        )


# pylint: disable=too-many-ancestors
class ExportXliffView(PageBulkActionMixin, BulkActionView):
    """
    Bulk action for generating XLIFF files for translations
    """

    #: Whether only public translation should be exported
    only_public = False
    #: Whether the view requires change permissions
    require_change_permission = False

    def post(self, request, *args, **kwargs):
        r"""
        Function for handling a XLIFF export request for pages.
        The pages get extracted from request.GET attribute and the request is forwarded to :func:`~integreat_cms.xliff.utils.pages_to_xliff_file`

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The redirect
        :rtype: ~django.http.HttpResponseRedirect
        """
        target_language = get_object_or_404(
            self.request.region.language_tree_nodes,
            language__slug=kwargs.get("language_slug"),
            parent__isnull=False,
        ).language

        xliff_file_url = pages_to_xliff_file(
            request,
            self.get_queryset(),
            [target_language],
            only_public=self.only_public,
        )
        if xliff_file_url:
            # Insert link with automatic download into success message
            messages.success(
                request,
                __(
                    _(
                        "XLIFF file with published pages only for translation to {} successfully created."
                    ).format(target_language)
                    if self.only_public
                    else _(
                        "XLIFF file with unpublished and published pages for translation to {} successfully created."
                    ).format(target_language),
                    _(
                        "If the download does not start automatically, please click {}here{}."
                    ).format(
                        f"<a data-auto-download href='{xliff_file_url}' class='font-bold underline hover:no-underline' download>",
                        "</a>",
                    ),
                ),
            )

        # Let the base view handle the redirect
        return super().post(request, *args, **kwargs)


class ExportMultiLanguageXliffView(PageBulkActionMixin, BulkActionView):
    """
    Bulk action for generating XLIFF files for translations in multiple languages.
    """

    #: Whether the view requires change permissions
    require_change_permission = False

    def post(self, request, *args, **kwargs):
        r"""
        Function for handling a XLIFF export request for pages and multiple languages.
        The pages get extracted from request.GET attribute and the request is forwarded to :func:`~integreat_cms.xliff.utils.pages_to_xliff_file`

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The redirect
        :rtype: ~django.http.HttpResponseRedirect
        """

        target_languages = [
            language
            for language in self.request.region.active_languages
            if language.slug in self.request.POST.getlist("selected_language_slugs[]")
        ]

        xliff_file_url = pages_to_xliff_file(
            request, self.get_queryset(), target_languages, only_public=False
        )
        if xliff_file_url:
            # Insert link with automatic download into success message
            messages.success(
                request,
                __(
                    _(
                        "XLIFF file for translation to selected languages successfully created."
                    ),
                    _(
                        "If the download does not start automatically, please click {}here{}."
                    ).format(
                        f"<a data-auto-download href='{xliff_file_url}' class='font-bold underline hover:no-underline' download>",
                        "</a>",
                    ),
                ),
            )

        # Let the base view handle the redirect
        return super().post(request, *args, **kwargs)



"""
This module contains labels for the choices regarding the position of embedding live content
"""
from django.utils.translation import ugettext_lazy as _


CHOICES = (
    (True, _("Embed mirrored page before this page")),
    (False, _("Embed mirrored page after this page")),
)



"""
This module contains view actions for user objects.
"""
import logging
from django.contrib import messages
from django.contrib.auth import get_user_model
from django.shortcuts import get_object_or_404, redirect
from django.utils.translation import ugettext as _
from django.views.decorators.http import require_POST

from ...decorators import permission_required
from ...utils.welcome_mail_utils import send_welcome_mail

logger = logging.getLogger(__name__)


@require_POST
@permission_required("cms.delete_user")
def delete_user(request, user_id):
    """
    This view deletes a user

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param user_id: The id of the user which should be deleted
    :type user_id: int

    :return: A redirection to user list
    :rtype: ~django.http.HttpResponseRedirect
    """

    user = get_object_or_404(get_user_model(), id=user_id)
    logger.info("%r deleted %r", request.user, user)
    user.delete()

    messages.success(request, _("User was successfully deleted"))

    return redirect("users")


@require_POST
@permission_required("cms.change_user")
def resend_activation_link(request, user_id):
    """
    Resends an activation link to an user

    :param request: The current request
    :type request: ~django.http.HttpResponse

    :param user_id: users id to send the activation link
    :type user_id: int

    :return: A redirection to region user list
    :rtype: ~django.http.HttpResponseRedirect
    """
    user = get_object_or_404(get_user_model(), id=user_id)
    send_welcome_mail(request, user, activation=True)
    return redirect("users")



from django.contrib.auth.models import Group

from ..custom_model_form import CustomModelForm


class GroupForm(CustomModelForm):
    """
    Form for creating and modifying user group objects
    """

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = Group
        #: The fields of the model which should be handled by this form
        fields = ["permissions"]

    def __init__(self, **kwargs):

        # Instantiate CustomModelForm
        super().__init__(**kwargs)
        # TODO: Derive size from view height (fill complete available space with select field)
        self.fields["permissions"].widget.attrs["size"] = "20"



import asyncio
import logging
import re

from datetime import date, datetime
from itertools import cycle
from urllib.parse import urlencode

import aiohttp

from django.conf import settings
from django.utils.translation import ugettext as _


from ..constants import colors, matomo_periods

logger = logging.getLogger(__name__)


class MatomoException(Exception):
    """
    Custom Exception class for errors during interaction with Matomo
    """


class MatomoApiManager:
    """
    This class helps to interact with Matomo API.
    There are three functions which can be used publicly:

    * :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.get_matomo_id`: Retrieve the Matomo ID belonging to the given Matomo access token
    * :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.get_total_visits`: Retrieve the total visits for the current region
    * :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.get_visits_per_language`: Retrieve the visits for the current region by language
    """

    #: Matomo API-key
    matomo_token = None
    #: Matomo ID
    matomo_id = None
    #: The active languages
    languages = None

    def __init__(self, region):
        """
        Constructor initializes the class variables

        :param region: The region this Matomo API Manager connects to
        :type region: ~integreat_cms.cms.models.regions.region.Region
        """
        self.region_name = region.name
        self.matomo_token = region.matomo_token
        self.matomo_id = region.matomo_id
        self.languages = region.active_languages

    async def fetch(self, session, **kwargs):
        r"""
        Uses :meth:`aiohttp.ClientSession.get` to perform an asynchronous GET request to the Matomo API.

        :param session: The session object which is used for the request
        :type session: aiohttp.ClientSession

        :param \**kwargs: The parameters which are passed to the Matomo API
        :type \**kwargs: dict

        :raises ~integreat_cms.cms.utils.matomo_api_manager.MatomoException: When a :class:`~aiohttp.ClientError` was raised during a
                                                               Matomo API request

        :return: The parsed :mod:`json` result
        :rtype: dict
        """
        # The default get parameters for all requests
        query_params = {
            "format": "JSON",
            "module": "API",
            "token_auth": self.matomo_token,
        }
        # Update with the custom params for this request
        query_params.update(kwargs)

        url = f"{settings.MATOMO_URL}/?{urlencode(query_params)}"
        logger.debug(
            "Requesting %r: %s",
            query_params.get("method"),
            # Mask auth token in log
            re.sub(r"&token_auth=[^&]+", "&token_auth=********", url),
        )
        try:
            async with session.get(url) as response:
                response_data = await response.json()
                if (
                    isinstance(response_data, dict)
                    and response_data.get("result") == "error"
                ):
                    raise MatomoException(response_data["message"])
                return response_data
        except aiohttp.ClientError as e:
            raise MatomoException(str(e)) from e

    async def get_matomo_id_async(self, **query_params):
        r"""
        Async wrapper to fetch the Matomo ID with :mod:`aiohttp`.
        Opens a :class:`~aiohttp.ClientSession` and calls :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.fetch`.
        Called from :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.get_matomo_id`.

        :param \**query_params: The parameters which are passed to the Matomo API
        :type \**query_params: dict

        :raises ~integreat_cms.cms.utils.matomo_api_manager.MatomoException: When a :class:`~aiohttp.ClientError` was raised during a
                                                               Matomo API request

        :return: The parsed :mod:`json` result
        :rtype: list
        """
        async with aiohttp.ClientSession() as session:
            return await self.fetch(session, **query_params)

    def get_matomo_id(self, token_auth):
        """
        Returns the matomo website id based on the provided authentication key.

        :param token_auth: The Matomo authentication token which should be used
        :type token_auth: str

        :raises ~integreat_cms.cms.utils.matomo_api_manager.MatomoException: When a :class:`~aiohttp.ClientError` was raised during a
                                                               Matomo API request or the access token is not correct

        :return: ID of the connected Matomo instance
        :rtype: int
        """
        # Initialize async event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        # Execute async request to Matomo API
        response = loop.run_until_complete(
            self.get_matomo_id_async(
                token_auth=token_auth,
                method="SitesManager.getSitesIdWithAtLeastViewAccess",
            )
        )

        try:
            return response[0]
        except IndexError as e:
            # If no id is returned, there is no user with the given access token
            raise MatomoException(
                f"The access token for {self.region_name} is not correct."
            ) from e

    async def get_total_visits_async(self, query_params):
        """
        Async wrapper to fetch the total visits with :mod:`aiohttp`.
        Opens a :class:`~aiohttp.ClientSession` and calls :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.fetch`.
        Called from :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.get_total_visits`.

        :param query_params: The parameters which are passed to the Matomo API
        :type query_params: dict

        :raises ~integreat_cms.cms.utils.matomo_api_manager.MatomoException: When a :class:`~aiohttp.ClientError` was raised during a
                                                               Matomo API request

        :return: The parsed :mod:`json` result
        :rtype: dict
        """
        async with aiohttp.ClientSession() as session:
            return await self.fetch(
                session,
                **query_params,
            )

    def get_total_visits(self, start_date, end_date, period=matomo_periods.DAY):
        """
        Returns the total calls within a time range for all languages.

        :param start_date: Start date
        :type start_date: ~datetime.date

        :param end_date: End date
        :type end_date: ~datetime.date

        :param period: The period (one of :attr:`~integreat_cms.cms.constants.matomo_periods.CHOICES` -
                       defaults to :attr:`~integreat_cms.cms.constants.matomo_periods.DAY`)
        :type period: str

        :raises ~integreat_cms.cms.utils.matomo_api_manager.MatomoException: When a :class:`~aiohttp.ClientError` was raised during a
                                                               Matomo API request

        :return: The total visits in the ChartData format expected by ChartJs
        :rtype: dict
        """
        query_params = {
            "date": f"{start_date},{end_date}",
            "idSite": self.matomo_id,
            "method": "VisitsSummary.getActions",
            "period": period,
        }

        # Initialize async event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        # Execute async request to Matomo API
        dataset = loop.run_until_complete(self.get_total_visits_async(query_params))

        return {
            # Send original labels for usage in the CSV export (convert to list because type dict_keys is not JSON-serializable)
            "exportLabels": list(dataset.keys()),
            # Return the data in the ChartData format expected by ChartJs
            "chartData": {
                # Make labels more readable
                "labels": self.simplify_date_labels(dataset.keys(), period),
                "datasets": [
                    {
                        "label": _("All languages"),
                        "borderColor": colors.DEFAULT,
                        "data": list(dataset.values()),
                    }
                ],
            },
        }

    async def get_visits_per_language_async(self, loop, query_params, languages):
        """
        Async wrapper to fetch the total visits with :mod:`aiohttp`.
        Opens a :class:`~aiohttp.ClientSession`, creates a :class:`~asyncio.Task` for each language to call
        :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.fetch` and waits for all tasks to finish with
        :func:`~asyncio.gather`.
        The returned list of gathered results has the correct order in which the tasks were created (at first the
        ordered list of languages and the last element is the task for the total visits).
        Called from :func:`~integreat_cms.cms.utils.matomo_api_manager.MatomoApiManager.get_visits_per_language`.

        :param loop: The asyncio event loop
        :type loop: asyncio.AbstractEventLoop

        :param query_params: The parameters which are passed to the Matomo API
        :type query_params: dict

        :param languages: The list of languages which should be retrieved
        :type languages: list [ ~integreat_cms.cms.models.languages.language.Language ]

        :raises ~integreat_cms.cms.utils.matomo_api_manager.MatomoException: When a :class:`~aiohttp.ClientError` was raised during a
                                                               Matomo API request

        :return: The list of gathered results
        :rtype: list
        """
        async with aiohttp.ClientSession() as session:
            # Create tasks for visits by language
            tasks = [
                loop.create_task(
                    self.fetch(
                        session,
                        **query_params,
                        segment=f"pageUrl=@/{language.slug}/wp-json/;pageUrl!@/pages/",
                    )
                )
                for language in languages
            ]
            # Create separate task to gather offline download hits
            tasks.append(
                loop.create_task(
                    self.fetch(session, **query_params, segment="pageUrl=@/pages/"),
                )
            )
            # Create task for total visits in all languages
            tasks.append(
                loop.create_task(
                    self.fetch(
                        session,
                        **query_params,
                    )
                )
            )
            # Wait for all tasks to finish and collect the results
            # (the results are sorted in the order the tasks were created)
            return await asyncio.gather(*tasks)

    def get_visits_per_language(self, start_date, end_date, period):
        """
        Returns the total unique visitors in a timerange as defined in period

        :param start_date: Start date
        :type start_date: ~datetime.date

        :param end_date: End date
        :type end_date: ~datetime.date

        :param period: The period (one of :attr:`~integreat_cms.cms.constants.matomo_periods.CHOICES`)
        :type period: str

        :raises ~integreat_cms.cms.utils.matomo_api_manager.MatomoException: When a :class:`~aiohttp.ClientError` was raised during a
                                                               Matomo API request

        :return: The visits per language in the ChartData format expected by ChartJs
        :rtype: dict
        """
        query_params = {
            "date": f"{start_date},{end_date}",
            "expanded": "1",
            "filter_limit": "-1",
            "format_metrics": "1",
            "idSite": self.matomo_id,
            "method": "VisitsSummary.getActions",
            "period": period,
        }
        logger.debug(
            "Query params: %r",
            query_params,
        )
        # Convert languages to a list to force an evaluation in the sync function
        # (in Django, database queries cannot be executed in async functions without more ado)
        languages = list(self.languages)
        # Convert colors to cycle to make sure it doesn't run out of elements if there are more languages than colors
        color_cycle = cycle(colors.CHOICES)

        # Initialize async event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        # Execute async request to Matomo API
        logger.debug("Fetching visits for languages %r asynchronously.", languages)
        datasets = loop.run_until_complete(
            self.get_visits_per_language_async(loop, query_params, languages)
        )
        logger.debug("All asynchronous fetching tasks have finished.")
        # The last dataset contains the total visits
        total_visits = datasets.pop()
        # Get the separately created datasets for offline downloads
        offline_downloads = datasets.pop()

        return {
            # Send original labels for usage in the CSV export (convert to list because type dict_keys is not JSON-serializable)
            "exportLabels": list(total_visits.keys()),
            # Return the data in the ChartData format expected by ChartJs
            "chartData": {
                # Make labels more readable
                "labels": self.simplify_date_labels(total_visits.keys(), period),
                "datasets":
                # The datasets for the visits by language
                [
                    {
                        "label": language.translated_name,
                        "borderColor": next(color_cycle),
                        "data": list(dataset.values()),
                    }
                    # zip aggregates two lists into tuples, e.g. zip([1,2,3], [4,5,6])=[(1,4), (2,5), (3,6)]
                    # In this case, it matches the languages to their respective dataset (because the datasets are ordered)
                    for language, dataset in zip(languages, datasets)
                ]
                # The dataset for offline downloads
                + [
                    {
                        "label": _("Offline Downloads"),
                        "borderColor": next(color_cycle),
                        "data": list(offline_downloads.values()),
                    }
                ]
                # The dataset for total visits
                + [
                    {
                        "label": _("All languages"),
                        "borderColor": colors.DEFAULT,
                        "data": list(total_visits.values()),
                    }
                ],
            },
        }

    @staticmethod
    def simplify_date_labels(date_labels, period):
        """
        Convert the dates returned by Matomo to more readable labels

        :param date_labels: The date labels returned by Matomo
        :type date_labels: list [ str ]

        :param period: The period of the labels (determines the format)
        :type period: str

        :return: The readable labels
        :rtype: list [ str ]
        """
        simplified_date_labels = []
        if period == matomo_periods.DAY:
            # Convert string labels to date objects (the format for daily period is the iso format YYYY-MM-DD)
            date_objects = [
                date.fromisoformat(date_label) for date_label in date_labels
            ]
            # Convert date objects to more readable labels
            if date.today().year == date_objects[0].year:
                # If the first label is in the current year, omit the year for all dates
                simplified_date_labels = [
                    date_obj.strftime("%d.%m.") for date_obj in date_objects
                ]
            else:
                # Else, include the year
                simplified_date_labels = [
                    date_obj.strftime("%d.%m.%Y") for date_obj in date_objects
                ]
        elif period == matomo_periods.WEEK:
            # Convert string labels to date objects (the format for weekly period is YYYY-MM-DD,YYYY-MM-DD)
            date_objects = [
                datetime.strptime(date_label.split(",")[0], "%Y-%m-%d").date()
                for date_label in date_labels
            ]
            # Convert date objects to more readable labels
            if date.today().year == date_objects[0].year:
                # If the first label is in the current year, omit the year for all dates
                simplified_date_labels = [
                    _("CW") + date_obj.strftime(" %W") for date_obj in date_objects
                ]
            else:
                # Else, include the year
                simplified_date_labels = [
                    date_obj.strftime("%Y ") + _("CW") + date_obj.strftime(" %W")
                    for date_obj in date_objects
                ]
        elif period == matomo_periods.MONTH:
            # Convert string labels to date objects (the format for monthly period is YYYY-MM)
            date_objects = [
                datetime.strptime(date_label, "%Y-%m").date()
                for date_label in date_labels
            ]
            # Convert date objects to more readable labels
            if date.today().year == date_objects[0].year:
                # If the first label is in the current year, omit the year for all dates
                simplified_date_labels = [
                    _(date_obj.strftime("%B")) for date_obj in date_objects
                ]
            else:
                # Else, include the year
                simplified_date_labels = [
                    _(date_obj.strftime("%B")) + date_obj.strftime(" %Y")
                    for date_obj in date_objects
                ]
        else:
            # This means the period is "year" (convert to list because type dict_keys is not JSON-serializable)
            simplified_date_labels = list(date_labels)
        return simplified_date_labels



"""
This module contains all constants representing the rating of the :class:`~integreat_cms.cms.models.feedback.feedback.Feedback` model.
"""
from django.utils.translation import ugettext_lazy as _


#: Positive
POSITIVE = True
#: Negative
NEGATIVE = False
#: Not stated
NOT_STATED = None
#: Not stated as string (used in the filter form)
NOT_STATED_STR = ""

#: Choices to use these constants in a database field
CHOICES = (
    (POSITIVE, _("Positive")),
    (NEGATIVE, _("Negative")),
    (NOT_STATED, _("Not stated")),
)

#: Choices for the filter form (required because ``None`` is always converted to ``""``)
FILTER_CHOICES = (
    (POSITIVE, _("Positive")),
    (NEGATIVE, _("Negative")),
    (NOT_STATED_STR, _("Not stated")),
)

#: Initial filter options
INITIAL = [key for (key, val) in FILTER_CHOICES]



"""
This is a collection of tags and filters for :class:`~integreat_cms.cms.models.pages.page.Page` objects.
"""
from django import template

register = template.Library()


@register.simple_tag
def get_depth_in(node, pageset):
    """
    This tag returns the depth of node whithin the tree/pages in pageset.

    :param node : the page
    :type node : ~integreat_cms.cms.models.pages.page.Page

    :param pageset: The pages (all pages or pages chosen by filter)
    :type pageset: list [ ~integreat_cms.cms.models.pages.page.Page ]

    :return: the depth of node whithin the tree/pages in pageset
    :rtype: int
    """
    if not node.parent in pageset:
        return 0
    return node.depth - get_highest_anscentor_in(node, pageset).depth


def get_highest_anscentor_in(node, pageset):
    """
    This tag returns the highest (farthest) anscestor of node whithin the tree/pages in pageset.

    :param node : the page
    :type node : ~integreat_cms.cms.models.pages.page.Page

    :param pageset: The pages (all pages or pages chosen by filter)
    :type pageset: list [ ~integreat_cms.cms.models.pages.page.Page ]

    :return: the highest (farthest) anscestor of node whithin the tree/pages in pageset
    :rtype:  ~integreat_cms.cms.models.pages.page.Page
    """
    if node.parent in pageset:
        return get_highest_anscentor_in(node.parent, pageset)
    return node



"""
This module contains all string representations of weekdays, used by :class:`~integreat_cms.cms.models.events.event.Event`.
"""
from django.utils.translation import ugettext_lazy as _


#: Monday
MONDAY = 0
#: Tuesday
TUESDAY = 1
#: Wednesday
WEDNESDAY = 2
#: Thursday
THURSDAY = 3
#: Friday
FRIDAY = 4
#: Saturday
SATURDAY = 5
#: Sunday
SUNDAY = 6

#: Choices to use these constants in a database field
CHOICES = (
    (MONDAY, _("Monday")),
    (TUESDAY, _("Tuesday")),
    (WEDNESDAY, _("Wednesday")),
    (THURSDAY, _("Thursday")),
    (FRIDAY, _("Friday")),
    (SATURDAY, _("Saturday")),
    (SUNDAY, _("Sunday")),
)



import logging

from ...models import POITranslation
from ..custom_content_model_form import CustomContentModelForm


logger = logging.getLogger(__name__)


class POITranslationForm(CustomContentModelForm):
    """
    Form for creating and modifying POI translation objects
    """

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = POITranslation
        #: The fields of the model which should be handled by this form
        fields = [
            "title",
            "short_description",
            "status",
            "content",
            "slug",
            "minor_edit",
        ]

    def __init__(self, **kwargs):
        r"""
        Initialize POI translation form

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict
        """

        # Pop kwarg to make sure the super class does not get this param
        default_language_title = kwargs.pop("default_language_title", None)

        # Instantiate CustomContentModelForm
        super().__init__(**kwargs)

        if default_language_title:
            self.fields["title"].initial = default_language_title



from datetime import datetime, time, date
from dateutil.rrule import weekday, rrule

from django.db import models
from django.db.models import Q
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.utils import timezone

from ...constants import frequency, status
from ...utils.slug_utils import generate_unique_slug
from ..abstract_content_model import AbstractContentModel, ContentQuerySet
from ..media.media_file import MediaFile
from ..pois.poi import POI
from .event_translation import EventTranslation
from .recurrence_rule import RecurrenceRule


class EventQuerySet(ContentQuerySet):
    """
    Custom QuerySet to facilitate the filtering by date while taking recurring events into account.
    """

    def filter_upcoming(self, from_date=None):
        """
        Filter all events that take place after the given date. This is, per definition, if at least one of the
        following conditions is true:

            * The end date of the event is the given date or later
            * The event is indefinitely recurring
            * The event is recurring and the recurrence end date is the given date or later

        :param from_date: The date which should be used for filtering, defaults to the current date
        :type from_date: datetime.date

        :return: The Queryset of events after the given date
        :rtype: ~integreat_cms.cms.models.events.event.EventQuerySet
        """
        from_date = from_date or timezone.now().date()
        return self.filter(
            Q(end__gte=from_date)
            | Q(
                recurrence_rule__isnull=False,
                recurrence_rule__recurrence_end_date__isnull=True,
            )
            | Q(
                recurrence_rule__isnull=False,
                recurrence_rule__recurrence_end_date__gte=from_date,
            )
        )

    def filter_completed(self, to_date=None):
        """
        Filter all events that are not ongoing and don't have any occurrences in the future. This is, per definition, if
        at least one of the following conditions is true:

            * The event is non-recurring and the end date of the event is before the given date
            * The event is recurring and the recurrence end date is before the given date

        :param to_date: The date which should be used for filtering, defaults to the current date
        :type to_date: datetime.date

        :return: The Queryset of events before the given date
        :rtype: ~integreat_cms.cms.models.events.event.EventQuerySet
        """
        to_date = to_date or timezone.now().date()
        return self.filter(
            Q(recurrence_rule__isnull=True, end__lt=to_date)
            | Q(
                recurrence_rule__isnull=False,
                recurrence_rule__recurrence_end_date__lt=to_date,
            )
        )


class Event(AbstractContentModel):
    """
    Data model representing an event.
    Can be directly imported from :mod:`~integreat_cms.cms.models`.
    """

    location = models.ForeignKey(
        POI,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name=_("location"),
    )
    start = models.DateTimeField(verbose_name=_("start"))
    end = models.DateTimeField(verbose_name=_("end"))
    #: If the event is recurring, the recurrence rule contains all necessary information on the frequency, interval etc.
    #: which is needed to calculate the single instances of a recurring event
    recurrence_rule = models.OneToOneField(
        RecurrenceRule,
        null=True,
        on_delete=models.SET_NULL,
        related_name="event",
        verbose_name=_("recurrence rule"),
    )
    icon = models.ForeignKey(
        MediaFile,
        verbose_name=_("icon"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    archived = models.BooleanField(default=False, verbose_name=_("archived"))

    #: The default manager
    objects = EventQuerySet.as_manager()

    @property
    def fallback_translations_enabled(self):
        """
        Whether translations should be returned in the default language if they do not exist

        :return: Whether fallback translations are enabled
        :rtype: bool
        """
        return self.region.fallback_translations_enabled

    @staticmethod
    def get_translation_model():
        """
        Returns the translation model of this content model

        :return: The class of translations
        :rtype: type
        """
        return EventTranslation

    @cached_property
    def is_recurring(self):
        """
        This property checks if the event has a recurrence rule and thereby determines, whether the event is recurring.

        :return: Whether the event is recurring or not
        :rtype: bool
        """
        return bool(self.recurrence_rule)

    @cached_property
    def is_all_day(self):
        """
        This property checks whether an event takes place the whole day by checking if start time is minimal and end
        time is maximal.

        :return: Whether event takes place all day
        :rtype: bool
        """
        return (
            self.start_local.time() == time.min
            and self.end_local.time() == time.max.replace(second=0, microsecond=0)
        )

    @cached_property
    def timezone(self):
        """
        The timezone of this event's region

        :return: The timezone of this event
        :rtype: str
        """
        return self.region.timezone

    @cached_property
    def start_local(self):
        """
        Convert the start to local time

        :return: The start of the event in local time
        :rtype: datetime.datetime
        """
        timezone.activate(self.timezone)
        return timezone.localtime(self.start)

    @cached_property
    def end_local(self):
        """
        Convert the end to local time

        :return: The end of the event in local time
        :rtype: datetime.datetime
        """
        # Activate the timezone of the event's region
        timezone.activate(self.timezone)
        return timezone.localtime(self.end)

    @cached_property
    def has_location(self):
        """
        This property checks whether the event has a physical location (:class:`~integreat_cms.cms.models.pois.poi.POI`).

        :return: Whether event has a physical location
        :rtype: bool
        """
        return bool(self.location)

    def get_occurrences(self, start, end):
        """
        Get occurrences of the event that overlap with ``[start, end]``.
        Expects ``start < end``.

        :param start: the begin of the requested interval.
        :type start: ~datetime.datetime

        :param end: the end of the requested interval.
        :type end: ~datetime.datetime

        :return: start datetimes of occurrences of the event that are in the given timeframe
        :rtype: list [ ~datetime.datetime ]
        """
        event_start = self.start
        event_end = self.end
        event_span = event_end - event_start
        recurrence = self.recurrence_rule
        if recurrence is not None:
            until = min(
                end,
                datetime.combine(
                    recurrence.recurrence_end_date
                    if recurrence.recurrence_end_date
                    else date.max,
                    time.max,
                ),
            )
            if recurrence.frequency in (frequency.DAILY, frequency.YEARLY):
                occurrences = rrule(
                    recurrence.frequency,
                    dtstart=event_start,
                    interval=recurrence.interval,
                    until=until,
                )
            elif recurrence.frequency == frequency.WEEKLY:
                occurrences = rrule(
                    recurrence.frequency,
                    dtstart=event_start,
                    interval=recurrence.interval,
                    byweekday=recurrence.weekdays_for_weekly,
                    until=until,
                )
            else:
                occurrences = rrule(
                    recurrence.frequency,
                    dtstart=event_start,
                    interval=recurrence.interval,
                    byweekday=weekday(
                        recurrence.weekday_for_monthly, recurrence.week_for_monthly
                    ),
                    until=until,
                )
            return [
                x
                for x in occurrences
                if start <= x <= end or start <= x + event_span <= end
            ]
        return (
            [event_start]
            if start <= event_start <= end or start <= event_end <= end
            else []
        )

    def copy(self, user):
        """
        This method creates a copy of this event and all of its translations.
        This method saves the new event.

        :param user: The user who initiated this copy
        :type user: ~django.contrib.auth.models.User

        :return: A copy of this event
        :rtype: ~integreat_cms.cms.models.events.event.Event
        """
        # save all translations on the original object, so that they can be copied later
        translations = list(self.translations.all())

        # Clear the own recurrence rule.
        # If the own recurrence rule would not be cleared, django would throw an
        # error that the recurrence rule is not unique (because it would belong to both
        # the cloned and the new object)
        recurrence_rule = self.recurrence_rule

        if recurrence_rule:
            # copy the recurrence rule, if it exists
            recurrence_rule.pk = None
            recurrence_rule.save()
            self.recurrence_rule = recurrence_rule

        # create the copied event
        self.pk = None
        self.save()

        copy_translation = _("copy")
        # Create new translations for this event
        for translation in translations:
            translation.pk = None
            translation.event = self
            translation.status = status.DRAFT
            translation.title = f"{translation.title} ({copy_translation})"
            translation.slug = generate_unique_slug(
                **{
                    "slug": translation.slug,
                    "manager": type(translation).objects,
                    "object_instance": translation,
                    "foreign_model": "event",
                    "instance": self,
                    "region": self.region,
                    "language": translation.language,
                }
            )
            translation.currently_in_translation = False
            translation.creator = user
            translation.save()

        return self

    class Meta:
        #: The verbose name of the model
        verbose_name = _("event")
        #: The plural verbose name of the model
        verbose_name_plural = _("events")
        #: The name that will be used by default for the relation from a related object back to this one
        default_related_name = "events"
        #: The fields which are used to sort the returned objects of a QuerySet
        ordering = ["start"]
        #: The default permissions for this model
        default_permissions = ("change", "delete", "view")
        #: The custom permissions for this model
        permissions = (("publish_event", "Can publish events"),)



"""
Configuration of DeepL API app
"""
import os
import sys
import logging

from deepl.exceptions import DeepLException

from django.apps import AppConfig
from django.conf import settings

from .utils import DeepLApi

logger = logging.getLogger(__name__)


class DeepLApiConfig(AppConfig):
    """
    DeepL API config inheriting the django AppConfig
    """

    name = "integreat_cms.deepl_api"
    supported_source_languages = []
    supported_target_languages = []

    def ready(self):
        """
        Checking if API is available
        """
        # Only check availability if running a server
        if "runserver" in sys.argv or "APACHE_PID_FILE" in os.environ:
            if settings.DEEPL_ENABLED:
                try:
                    deepl = DeepLApi()
                    self.supported_source_languages = [
                        source_language.code.lower()
                        for source_language in deepl.translator.get_source_languages()
                    ]
                    logger.debug(
                        "Supported source languages by DeepL: %r",
                        self.supported_source_languages,
                    )
                    assert self.supported_source_languages
                    self.supported_target_languages = [
                        target_languages.code.lower()[:2]
                        for target_languages in deepl.translator.get_target_languages()
                    ]
                    logger.debug(
                        "Supported target languages by DeepL: %r",
                        self.supported_target_languages,
                    )
                    assert self.supported_target_languages
                    # pylint: disable=protected-access
                    logger.info(
                        "DeepL API is available at: %r", deepl.translator._server_url
                    )
                except (DeepLException, AssertionError) as e:
                    logger.exception(e)
                    logger.error(
                        "DeepL API is unavailable. You won't be able to "
                        "automatically translate events and locations."
                    )
            else:
                logger.info("DeepL API is disabled.")



from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from .feedback import Feedback


class EventListFeedback(Feedback):
    """
    Database model representing feedback about the event list (e.g. missing events).
    """

    @property
    def object_name(self):
        """
        This property returns the name of the object this feedback comments on.

        :return: The name of the object this feedback refers to
        :rtype: str
        """
        return _("Event list")

    @cached_property
    def object_url(self):
        """
        This property returns the url to the object this feedback comments on.

        :return: The url to the referred object
        :rtype: str
        """
        return reverse(
            "events",
            kwargs={
                "region_slug": self.region.slug,
            },
        )

    @property
    def related_feedback(self):
        """
        This property returns all feedback entries which relate to the same object and have the same is_technical value.

        :return: The queryset of related feedback
        :rtype: ~django.db.models.query.QuerySet [ ~integreat_cms.cms.models.feedback.event_list_feedback.EventListFeedback ]
        """
        return EventListFeedback.objects.filter(
            region=self.region, language=self.language, is_technical=self.is_technical
        )

    class Meta:
        #: The verbose name of the model
        verbose_name = _("event list feedback")
        #: The plural verbose name of the model
        verbose_name_plural = _("event list feedback")
        #: The default permissions for this model
        default_permissions = ()



# Generated by Django 3.2.11 on 2022-05-18 00:05
from django.db import migrations, models

from ..constants import roles


# pylint: disable=unused-argument
def update_roles(apps, schema_editor):
    """
    Update the role definitions

    :param apps: The configuration of installed applications
    :type apps: ~django.apps.registry.Apps

    :param schema_editor: The database abstraction layer that creates actual SQL code
    :type schema_editor: ~django.db.backends.base.schema.BaseDatabaseSchemaEditor
    """
    # We can't import the Person model directly as it may be a newer
    # version than this migration expects. We use the historical version.
    Group = apps.get_model("auth", "Group")
    Permission = apps.get_model("auth", "Permission")
    Role = apps.get_model("cms", "Role")

    # Rename municipality team to service team
    municipality_team = Group.objects.get(name="MUNICIPALITY_TEAM")
    municipality_team.name = "SERVICE_TEAM"
    municipality_team.save()
    municipality_team.role.name = "SERVICE_TEAM"
    municipality_team.role.save()

    # Create author role
    author, _ = Group.objects.get_or_create(name="AUTHOR")
    Role.objects.get_or_create(name="AUTHOR", group=author)

    # Assign the correct permissions
    for role_name in dict(roles.CHOICES):
        group = Group.objects.filter(name=role_name).first()
        # Clear permissions
        group.permissions.clear()
        # Set permissions
        group.permissions.add(
            *Permission.objects.filter(codename__in=roles.PERMISSIONS[role_name])
        )


class Migration(migrations.Migration):
    """
    Migration file to update the role definitions
    """

    dependencies = [
        ("cms", "0024_region_locations_enabled"),
    ]

    operations = [
        migrations.RunPython(update_roles, migrations.RunPython.noop),
        migrations.AlterField(
            model_name="role",
            name="name",
            field=models.CharField(
                choices=[
                    ("MANAGEMENT", "Management"),
                    ("EDITOR", "Editor"),
                    ("AUTHOR", "Author"),
                    ("EVENT_MANAGER", "Event manager"),
                    ("SERVICE_TEAM", "Service team"),
                    ("CMS_TEAM", "CMS team"),
                    ("APP_TEAM", "App team"),
                    ("MARKETING_TEAM", "Marketing team"),
                ],
                max_length=50,
                verbose_name="name",
            ),
        ),
    ]



"""
A view representing an instance of a point of interest. POIs can be created or updated via this view.
"""
import logging

from django.conf import settings
from django.contrib import messages
from django.shortcuts import render, redirect
from django.utils.decorators import method_decorator
from django.utils.translation import ugettext as _
from django.views.generic import TemplateView

from ...constants import status
from ...decorators import permission_required
from ...forms import POIForm, POITranslationForm
from ...models import POI, POITranslation, Language
from ...utils.translation_utils import ugettext_many_lazy as __
from ..media.media_context_mixin import MediaContextMixin
from ..mixins import ContentEditLockMixin
from .poi_context_mixin import POIContextMixin

logger = logging.getLogger(__name__)


@method_decorator(permission_required("cms.view_poi"), name="dispatch")
@method_decorator(permission_required("cms.change_poi"), name="post")
class POIFormView(
    TemplateView, POIContextMixin, MediaContextMixin, ContentEditLockMixin
):
    """
    View for editing POIs
    """

    #: The template to render (see :class:`~django.views.generic.base.TemplateResponseMixin`)
    template_name = "pois/poi_form.html"
    #: The context dict passed to the template (see :class:`~django.views.generic.base.ContextMixin`)
    extra_context = {"current_menu_item": "pois_form"}
    #: The url name of the view to show if the user decides to go back (see :class:`~integreat_cms.cms.views.mixins.ContentEditLockMixin`)
    back_url_name = "pois"

    def get(self, request, *args, **kwargs):
        r"""
        Render :class:`~integreat_cms.cms.forms.pois.poi_form.POIForm` and :class:`~integreat_cms.cms.forms.pois.poi_translation_form.POITranslationForm`

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region
        language = Language.objects.get(slug=kwargs.get("language_slug"))

        # get poi and translation objects if they exist
        poi = POI.objects.filter(id=kwargs.get("poi_id")).first()
        poi_translation = POITranslation.objects.filter(
            poi=poi,
            language=language,
        ).first()

        if poi and poi.archived:
            disabled = True
            messages.warning(
                request, _("You cannot edit this location because it is archived.")
            )
        else:
            disabled = False

        poi_form = POIForm(instance=poi, disabled=disabled)
        poi_translation_form = POITranslationForm(
            instance=poi_translation,
            disabled=disabled,
            default_language_title=poi.default_translation.title if poi else None,
        )
        url_link = f"{settings.WEBAPP_URL}/{region.slug}/{language.slug}/{poi_translation_form.instance.url_infix}/"
        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "poi_form": poi_form,
                "poi_translation_form": poi_translation_form,
                "language": language,
                # Languages for tab view
                "languages": region.active_languages if poi else [language],
                "url_link": url_link,
                "translation_states": poi.translation_states if poi else [],
            },
        )

    # pylint: disable=too-many-branches,too-many-locals,unused-argument
    def post(self, request, *args, **kwargs):
        r"""
        Submit :class:`~integreat_cms.cms.forms.pois.poi_form.POIForm` and
        :class:`~integreat_cms.cms.forms.pois.poi_translation_form.POITranslationForm` and save :class:`~integreat_cms.cms.models.pois.poi.POI` and
        :class:`~integreat_cms.cms.models.pois.poi_translation.POITranslation` objects

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict

        :return: The rendered template response
        :rtype: ~django.template.response.TemplateResponse
        """

        region = request.region
        language = Language.objects.get(slug=kwargs.get("language_slug"))

        poi_instance = POI.objects.filter(id=kwargs.get("poi_id")).first()
        poi_translation_instance = POITranslation.objects.filter(
            poi=poi_instance,
            language=language,
        ).first()

        if poi_instance and poi_instance.archived:
            return redirect(
                "edit_poi",
                **{
                    "poi_id": poi_instance.id,
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )

        poi_form = POIForm(
            data=request.POST,
            files=request.FILES,
            instance=poi_instance,
            additional_instance_attributes={
                "region": region,
            },
        )
        poi_translation_form = POITranslationForm(
            data=request.POST,
            instance=poi_translation_instance,
            additional_instance_attributes={
                "creator": request.user,
                "language": language,
                "poi": poi_form.instance,
            },
            changed_by_user=request.user,
        )

        if not poi_form.is_valid() or not poi_translation_form.is_valid():
            # Add error messages
            poi_form.add_error_messages(request)
            poi_translation_form.add_error_messages(request)
        elif (
            poi_translation_form.instance.status == status.AUTO_SAVE
            and not poi_form.has_changed()
            and not poi_translation_form.has_changed()
        ):
            messages.info(request, _("No changes detected, autosave skipped"))
        else:
            # Save forms
            poi_translation_form.instance.poi = poi_form.save()
            poi_translation_form.save()
            # If any source translation changes to draft, set all depending translations/versions to draft
            if poi_translation_form.instance.status == status.DRAFT:
                language_tree_node = region.language_node_by_slug.get(language.slug)
                languages = [language] + [
                    node.language for node in language_tree_node.get_descendants()
                ]
                poi_translation_form.instance.poi.translations.filter(
                    language__in=languages
                ).update(status=status.DRAFT)

            # Add the success message and redirect to the edit page
            if not poi_instance:
                messages.success(
                    request,
                    _('Location "{}" was successfully created').format(
                        poi_translation_form.instance
                    ),
                )
            elif not poi_form.has_changed() and not poi_translation_form.has_changed():
                messages.info(request, _("No changes detected, but date refreshed"))
            else:
                # Add the success message
                poi_translation_form.add_success_message(request)
            # Show warning if nominatim result is more than 10km away from manually entered coordinates
            if poi_form.nominatim_distance_delta > 10:
                messages.warning(
                    request,
                    __(
                        _(
                            "The distance between the manually entered coordinates and the coordinates of the address is {}km."
                        ).format(poi_form.nominatim_distance_delta),
                        _("Please make sure the entered values are correct."),
                    ),
                )
            return redirect(
                "edit_poi",
                **{
                    "poi_id": poi_form.instance.id,
                    "region_slug": region.slug,
                    "language_slug": language.slug,
                },
            )
        url_link = f"{settings.WEBAPP_URL}/{region.slug}/{language.slug}/{poi_translation_form.instance.url_infix}/"
        return render(
            request,
            self.template_name,
            {
                **self.get_context_data(**kwargs),
                "poi_form": poi_form,
                "poi_translation_form": poi_translation_form,
                "language": language,
                # Languages for tab view
                "languages": region.active_languages if poi_instance else [language],
                "url_link": url_link,
                "translation_states": poi_instance.translation_states
                if poi_instance
                else [],
            },
        )



"""
This module contains all possible positions inside a MPTT tree model.
It is used to determine the relative position of nodes in relation to another node.
"""
from django.utils.translation import ugettext_lazy as _


#: First child of node
FIRST_CHILD = "first-child"
#: Last child of node
LAST_CHILD = "last-child"
#: Left neighbor of node
LEFT = "left"
#: Right neighbor of node
RIGHT = "right"
#: First sibling of node
FIRST_SIBLING = "first-sibling"
#: Last sibling of node
LAST_SIBLING = "last-sibling"

#: Choices to use these constants in a database field
CHOICES = (
    (FIRST_CHILD, _("First subpage of")),
    (LAST_CHILD, _("Last subpage of")),
    (LEFT, _("Left neighbor of")),
    (RIGHT, _("Right neighbor of")),
)



from ..custom_model_form import CustomModelForm
from ...models import Language


class LanguageForm(CustomModelForm):
    """
    Form for creating and modifying language objects
    """

    class Meta:
        """
        This class contains additional meta configuration of the form class, see the :class:`django.forms.ModelForm`
        for more information.
        """

        #: The model of this :class:`django.forms.ModelForm`
        model = Language
        #: The fields of the model which should be handled by this form
        fields = [
            "slug",
            "bcp47_tag",
            "english_name",
            "native_name",
            "text_direction",
            "table_of_contents",
            "primary_country_code",
            "secondary_country_code",
            "message_content_not_available",
            "message_partial_live_content_not_available",
        ]

    def __init__(self, *args, **kwargs):
        r"""
        Initialize language form

        :param \*args: The supplied arguments
        :type \*args: list

        :param \**kwargs: The supplied keyword arguments
        :type \**kwargs: dict
        """

        # Instantiate ModelForm
        super().__init__(*args, **kwargs)

        # Sort countries by translated name
        sorted_language_choices = sorted(
            self.fields["primary_country_code"].choices, key=lambda x: x[1]
        )
        self.fields["primary_country_code"].choices = sorted_language_choices
        self.fields["secondary_country_code"].choices = sorted_language_choices

        # Make left border rounded if no flag is selected yet
        if not self.instance.primary_country_code:
            self.fields["primary_country_code"].widget.attrs[
                "class"
            ] = "rounded-l border-l"
        if not self.instance.secondary_country_code:
            self.fields["secondary_country_code"].widget.attrs[
                "class"
            ] = "rounded-l border-l"



"""
This package contains all views related to the imprint
"""
from .imprint_form_view import ImprintFormView
from .imprint_revision_view import ImprintRevisionView
from .imprint_sbs_view import ImprintSideBySideView
from .imprint_actions import (
    delete_imprint,
    expand_imprint_translation_id,
)



import logging

from webauthn import generate_authentication_options, options_to_json
from webauthn.helpers import bytes_to_base64url
from webauthn.helpers.structs import PublicKeyCredentialDescriptor


from django.http.response import HttpResponse
from django.conf import settings
from django.contrib.auth import get_user_model
from django.http import JsonResponse
from django.utils.translation import ugettext as _
from django.views.generic import View

logger = logging.getLogger(__name__)


class MfaAssertView(View):
    """
    Generate challenge for multi factor authentication.
    If the user did not provide the first factor (password) or already authenticated with multiple factors, an error is
    returned.
    This AJAX view is called asynchronously by JavaScript.
    """

    def get(self, request):
        """

        :param request: The current request
        :type request: ~django.http.HttpResponse

        :return: The mfa challenge as JSON
        :rtype: ~django.http.JsonResponse
        """
        if "mfa_user_id" not in request.session:
            return JsonResponse(
                {"success": False, "error": _("You need to log in first")}, status=403
            )

        if request.user.is_authenticated:
            return JsonResponse(
                {"success": False, "error": _("You are already logged in.")}
            )
        user = get_user_model().objects.get(id=request.session["mfa_user_id"])

        webauthn_challenge = generate_authentication_options(
            rp_id=settings.HOSTNAME,
            allow_credentials=[
                PublicKeyCredentialDescriptor(id=key.key_id)
                for key in user.mfa_keys.all()
            ],
        )

        request.session["challenge"] = bytes_to_base64url(webauthn_challenge.challenge)

        # pylint: disable=http-response-with-content-type-json
        return HttpResponse(
            options_to_json(webauthn_challenge), content_type="application/json"
        )



import logging

from django.apps import AppConfig
from django.contrib.auth.signals import (
    user_logged_in,
    user_logged_out,
    user_login_failed,
)
from django.dispatch import receiver

logger = logging.getLogger(__name__)


class CmsConfig(AppConfig):
    """
    This class represents the Django-configuration of the backend.

    See :class:`django.apps.AppConfig` for more information.

    :param name: The name of the app
    :type name: str
    """

    name = "integreat_cms.cms"

    # pylint: disable=unused-import,import-outside-toplevel
    def ready(self):
        # Implicitly connect a signal handlers decorated with @receiver.
        from .signals import feedback_signals


authlog = logging.getLogger("auth")


# pylint: disable=unused-argument
@receiver(user_logged_in)
def user_logged_in_callback(sender, request, user, **kwargs):
    r"""
    Log a successful login event

    :param sender: The class of the user that just logged in.
    :type sender: type

    :param request: The current request
    :type request: ~django.http.HttpRequest

    :param user: The user instance that just logged in.
    :type user: ~django.contrib.auth.models.User

    :param \**kwargs: The supplied keyword arguments
    :type \**kwargs: dict
    """
    ip = request.META.get("REMOTE_ADDR")
    authlog.info("login user=%s, ip=%s", user, ip)


# pylint: disable=unused-argument
@receiver(user_logged_out)
def user_logged_out_callback(sender, request, user, **kwargs):
    r"""
    Log a logout event

    :param sender: The class of the user that just logged out or ``None`` if the user was not authenticated.
    :type sender: type

    :param request: The current request
    :type request: ~django.http.HttpRequest

    :param user: The user instance that just logged out or ``None`` if the user was not authenticated.
    :type user: ~django.contrib.auth.models.User

    :param \**kwargs: The supplied keyword arguments
    :type \**kwargs: dict
    """
    ip = request.META.get("REMOTE_ADDR")
    authlog.info("logout user=%s, ip=%s", user, ip)


# pylint: disable=unused-argument
@receiver(user_login_failed)
def user_login_failed_callback(sender, credentials, request, **kwargs):
    r"""
    Log a failed login event

    :param sender: The name of the module used for authentication.
    :type sender: str

    :param credentials: A dictionary of keyword arguments containing the user credentials that were passed to
                        :func:`~django.contrib.auth.authenticate`. Credentials matching a set of ‘sensitive’ patterns,
                        (including password) will not be sent in the clear as part of the signal.
    :type credentials: dict

    :param request: The current request
    :type request: ~django.http.HttpRequest

    :param \**kwargs: The supplied keyword arguments
    :type \**kwargs: dict
    """
    ip = request.META.get("REMOTE_ADDR")
    authlog.warning("login failed user=%s, ip=%s", credentials["username"], ip)



from urllib.parse import urlencode

import pytest

from django.urls import reverse


# pylint: disable=unused-argument,too-many-locals
@pytest.mark.django_db
def test_page_filters(load_test_data, admin_client):
    """
    Test whether duplicating regions works as expected

    :param load_test_data: The fixture providing the test data (see :meth:`~tests.conftest.load_test_data`)
    :type load_test_data: tuple

    :param admin_client: The fixture providing the logged in admin
    :type admin_client: :fixture:`admin_client`
    """
    page_tree = reverse(
        "pages", kwargs={"region_slug": "augsburg", "language_slug": "en"}
    )
    # Get with no data should result in the normal, unfiltered view with only root pages
    response = admin_client.get(page_tree)
    print(response.headers)
    assert response.status_code == 200
    # Root pages should be in the result
    assert "Welcome" in response.content.decode("utf-8")
    # Sub-pages should not be in the result
    assert "Trivia about Augsburg" not in response.content.decode("utf-8")
    assert "About the Integreat App Augsburg" not in response.content.decode("utf-8")
    # Searching for a page should return all results, independently of their level
    filter_params = {"query": "Augsburg"}
    response = admin_client.get(f"{page_tree}?{urlencode(filter_params)}")
    print(response.headers)
    assert response.status_code == 200
    # Pages without the query in the title should not be in the result
    assert "Welcome" not in response.content.decode("utf-8")
    # Pages with the query in the title should be in the result
    assert "Trivia about Augsburg" in response.content.decode("utf-8")
    assert "About the Integreat App Augsburg" in response.content.decode("utf-8")
    # Test filtering for the translation status
    filter_params = {"translation_status": ["OUTDATED", "MISSING"]}
    response = admin_client.get(f"{page_tree}?{urlencode(filter_params, True)}")
    print(f"{page_tree}?{urlencode(filter_params, True)}")
    print(response.headers)
    assert response.status_code == 200
    # Up-to-date pages should not be in the result
    assert "Welcome" not in response.content.decode("utf-8")
    # Missing and outdated pages should be in the result
    assert "Trivia about Augsburg" in response.content.decode("utf-8")
    assert "About the Integreat App Augsburg" in response.content.decode("utf-8")
    # Test filtering for the translation status and searching at the same time
    filter_params = {"query": "Trivia", "translation_status": ["OUTDATED", "MISSING"]}
    response = admin_client.get(f"{page_tree}?{urlencode(filter_params, True)}")
    print(response.headers)
    assert response.status_code == 200
    # Up-to-date pages should not be in the result
    assert "Welcome" not in response.content.decode("utf-8")
    # Missing and outdated pages with "Trivia" in the title should be in the result
    assert "Trivia about Augsburg" in response.content.decode("utf-8")
    # Missing and outdated pages without "Trivia" should not be in the result
    assert "About the Integreat App Augsburg" not in response.content.decode("utf-8")
    # Test filtering for the translation status and searching at the same time
    filter_params = {"query": "Welco", "translation_status": ["OUTDATED", "MISSING"]}
    response = admin_client.get(f"{page_tree}?{urlencode(filter_params, True)}")
    print(response.headers)
    assert response.status_code == 200
    # Up-to-date pages should not be in the result, even if they contain "Welco" in the title
    assert "Welcome" not in response.content.decode("utf-8")
    # Missing and outdated pages without "Welco" should not be in the result
    assert "Trivia about Augsburg" not in response.content.decode("utf-8")
    assert "About the Integreat App Augsburg" not in response.content.decode("utf-8")



"""
This module includes functions related to the locations/POIs API endpoint.
"""
from distutils.util import strtobool
from django.conf import settings
from django.http import JsonResponse
from django.utils import timezone
from ..decorators import json_response


def transform_poi(poi):
    """
    Function to create a JSON from a single poi object.

    :param poi: The poi object which should be converted
    :type poi: ~integreat_cms.cms.models.pois.poi.POI

    :return: data necessary for API
    :rtype: dict
    """
    if not poi:
        return {
            "id": None,
            "name": None,
            "address": None,
            "town": None,
            "state": None,
            "postcode": None,
            "region": None,
            "country": None,
            "latitude": None,
            "longitude": None,
        }
    return {
        "id": poi.id,
        "name": poi.default_translation.title,
        "address": poi.address,
        "town": poi.city,
        "state": None,
        "postcode": poi.postcode,
        "region": None,
        "country": poi.country,
        "latitude": poi.latitude,
        "longitude": poi.longitude,
    }


def transform_poi_translation(poi_translation):
    """
    Function to create a JSON from a single poi_translation object.

    :param poi_translation: The poi translation object which should be converted
    :type poi_translation: ~integreat_cms.cms.models.pois.poi_translation.POITranslation

    :return: data necessary for API
    :rtype: dict
    """

    poi = poi_translation.poi
    return {
        "id": poi_translation.id,
        "url": settings.BASE_URL + poi_translation.get_absolute_url(),
        "path": poi_translation.get_absolute_url(),
        "title": poi.default_translation.title,
        "modified_gmt": poi_translation.last_updated,  # deprecated field in the future
        "last_updated": timezone.localtime(poi_translation.last_updated),
        "excerpt": poi_translation.short_description,
        "content": poi_translation.content,
        "available_languages": poi_translation.available_languages_dict,
        "icon": poi.icon.url if poi.icon else None,
        "thumbnail": poi.icon.thumbnail_url if poi.icon else None,
        "website": poi.website if poi.website else None,
        "email": poi.email if poi.email else None,
        "phone_number": poi.phone_number if poi.phone_number else None,
        "location": transform_poi(poi),
        "hash": None,
    }


@json_response
# pylint: disable=unused-argument
def locations(request, region_slug, language_slug):
    """
    List all POIs of the region and transform result into JSON

    :param request: The current request
    :type request: ~django.http.HttpRequest

    :param region_slug: The slug of the requested region
    :type region_slug: str

    :param language_slug: The slug of the requested language
    :type language_slug: str

    :return: JSON object according to APIv3 locations endpoint definition
    :rtype: ~django.http.JsonResponse
    """
    region = request.region
    # Throw a 404 error when the language does not exist or is disabled
    region.get_language_or_404(language_slug, only_active=True)
    result = []
    pois = region.pois.prefetch_public_translations().filter(archived=False)
    if "on_map" in request.GET:
        try:
            location_on_map = strtobool(request.GET["on_map"])
        except ValueError as e:
            return JsonResponse({"error": str(e)}, status=400)
        pois = pois.filter(location_on_map=location_on_map)
    for poi in pois:
        translation = poi.get_public_translation(language_slug)
        if translation:
            result.append(transform_poi_translation(translation))

    return JsonResponse(
        result, safe=False
    )  # Turn off Safe-Mode to allow serializing arrays
